# /*##########################################################################
#
# Copyright (c) 2018-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module define ROIs for profile tools.

.. inheritance-diagram::
   silx.gui.plot.tools.profile.rois
   :top-classes: silx.gui.plot.tools.profile.core.ProfileRoiMixIn, silx.gui.plot.items.roi.RegionOfInterest
   :parts: 1
   :private-bases:
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "01/12/2020"

import numpy
import weakref
from concurrent.futures import CancelledError

from silx.gui import colors

from silx.gui.plot import items
from silx.gui.plot.items import roi as roi_items
from . import core
from silx.gui import utils
from .....utils.proxy import docstring


def _relabelAxes(plot, text):
    """Relabel {xlabel} and {ylabel} from this text using the corresponding
    plot axis label. If the axis label is empty, label it with "X" and "Y".

    :rtype: str
    """
    xLabel = plot.getXAxis().getLabel()
    if not xLabel:
        xLabel = "X"
    yLabel = plot.getYAxis().getLabel()
    if not yLabel:
        yLabel = "Y"
    return text.format(xlabel=xLabel, ylabel=yLabel)


def _lineProfileTitle(x0, y0, x1, y1):
    """Compute corresponding plot title

    This can be overridden to change title behavior.

    :param float x0: Profile start point X coord
    :param float y0: Profile start point Y coord
    :param float x1: Profile end point X coord
    :param float y1: Profile end point Y coord
    :return: Title to use
    :rtype: str
    """
    if x0 == x1:
        title = '{xlabel} = %g; {ylabel} = [%g, %g]' % (x0, y0, y1)
    elif y0 == y1:
        title = '{ylabel} = %g; {xlabel} = [%g, %g]' % (y0, x0, x1)
    else:
        m = (y1 - y0) / (x1 - x0)
        b = y0 - m * x0
        title = '{ylabel} = %g * {xlabel} %+g' % (m, b)

    return title


class _ImageProfileArea(items.Shape):
    """This shape displays the location of pixels used to compute the
    profile."""

    def __init__(self, parentRoi):
        items.Shape.__init__(self, "polygon")
        color = colors.rgba(parentRoi.getColor())
        self.setColor(color)
        self.setFill(True)
        self.setOverlay(True)
        self.setPoints([[0, 0], [0, 0]])  # Else it segfault

        self.__parentRoi = weakref.ref(parentRoi)
        parentRoi.sigItemChanged.connect(self._updateAreaProperty)
        parentRoi.sigRegionChanged.connect(self._updateArea)
        parentRoi.sigProfilePropertyChanged.connect(self._updateArea)
        parentRoi.sigPlotItemChanged.connect(self._updateArea)

    def getParentRoi(self):
        if self.__parentRoi is None:
            return None
        parentRoi = self.__parentRoi()
        if parentRoi is None:
            self.__parentRoi = None
        return parentRoi

    def _updateAreaProperty(self, event=None, checkVisibility=True):
        parentRoi = self.sender()
        if event == items.ItemChangedType.COLOR:
            parentRoi._updateItemProperty(event, parentRoi, self)
        elif event == items.ItemChangedType.VISIBLE:
            if self.getPlotItem() is not None:
                parentRoi._updateItemProperty(event, parentRoi, self)

    def _updateArea(self):
        roi = self.getParentRoi()
        item = roi.getPlotItem()
        if item is None:
            self.setVisible(False)
            return
        polygon = self._computePolygon(item)
        self.setVisible(True)
        polygon = numpy.array(polygon).T
        self.setLineStyle("--")
        self.setPoints(polygon, copy=False)

    def _computePolygon(self, item):
        if not isinstance(item, items.ImageBase):
            raise TypeError("Unexpected class %s" % type(item))

        currentData = item.getValueData(copy=False)

        roi = self.getParentRoi()
        origin = item.getOrigin()
        scale = item.getScale()
        _coords, _profile, area, _profileName, _xLabel = core.createProfile(
            roiInfo=roi._getRoiInfo(),
            currentData=currentData,
            origin=origin,
            scale=scale,
            lineWidth=roi.getProfileLineWidth(),
            method="none")
        return area


class _SliceProfileArea(items.Shape):
    """This shape displays the location a profile in a scatter.

    Each point used to compute the slice are linked together.
    """

    def __init__(self, parentRoi):
        items.Shape.__init__(self, "polygon")
        color = colors.rgba(parentRoi.getColor())
        self.setColor(color)
        self.setFill(True)
        self.setOverlay(True)
        self.setPoints([[0, 0], [0, 0]])  # Else it segfault

        self.__parentRoi = weakref.ref(parentRoi)
        parentRoi.sigItemChanged.connect(self._updateAreaProperty)
        parentRoi.sigRegionChanged.connect(self._updateArea)
        parentRoi.sigProfilePropertyChanged.connect(self._updateArea)
        parentRoi.sigPlotItemChanged.connect(self._updateArea)

    def getParentRoi(self):
        if self.__parentRoi is None:
            return None
        parentRoi = self.__parentRoi()
        if parentRoi is None:
            self.__parentRoi = None
        return parentRoi

    def _updateAreaProperty(self, event=None, checkVisibility=True):
        parentRoi = self.sender()
        if event == items.ItemChangedType.COLOR:
            parentRoi._updateItemProperty(event, parentRoi, self)
        elif event == items.ItemChangedType.VISIBLE:
            if self.getPlotItem() is not None:
                parentRoi._updateItemProperty(event, parentRoi, self)

    def _updateArea(self):
        roi = self.getParentRoi()
        item = roi.getPlotItem()
        if item is None:
            self.setVisible(False)
            return
        polylines = self._computePolylines(roi, item)
        if polylines is None:
            self.setVisible(False)
            return
        self.setVisible(True)
        self.setLineStyle("--")
        self.setPoints(polylines, copy=False)

    def _computePolylines(self, roi, item):
        slicing = roi._getSlice(item)
        if slicing is None:
            return None
        xx, yy, _values, _xx_error, _yy_error = item.getData(copy=False)
        xx, yy = xx[slicing], yy[slicing]
        polylines = numpy.array((xx, yy)).T
        if len(polylines) == 0:
            return None
        return polylines


class _DefaultImageProfileRoiMixIn(core.ProfileRoiMixIn):
    """Provide common behavior for silx default image profile ROI.
    """

    ITEM_KIND = items.ImageBase

    def __init__(self, parent=None):
        core.ProfileRoiMixIn.__init__(self, parent=parent)
        self.__method = "mean"
        self.__width = 1
        self.sigRegionChanged.connect(self.__regionChanged)
        self.sigPlotItemChanged.connect(self.__updateArea)
        self.__area = _ImageProfileArea(self)
        self.addItem(self.__area)

    def __regionChanged(self):
        self.invalidateProfile()
        self.__updateArea()

    def setProfileMethod(self, method):
        """
        :param str method: method to compute the profile. Can be 'mean' or 'sum'
        """
        if self.__method == method:
            return
        self.__method = method
        self.invalidateProperties()
        self.invalidateProfile()

    def getProfileMethod(self):
        return self.__method

    def setProfileLineWidth(self, width):
        if self.__width == width:
            return
        self.__width = width
        self.__updateArea()
        self.invalidateProperties()
        self.invalidateProfile()

    def getProfileLineWidth(self):
        return self.__width

    def __updateArea(self):
        plotItem = self.getPlotItem()
        if plotItem is None:
            self.setLineStyle("-")
        else:
            self.setLineStyle("--")

    def _getRoiInfo(self):
        """Wrapper to allow to reuse the previous Profile code.
    
        It would be good to remove it at one point.
        """
        if isinstance(self, roi_items.HorizontalLineROI):
            lineProjectionMode = 'X'
            y = self.getPosition()
            roiStart = (0, y)
            roiEnd = (1, y)
        elif isinstance(self, roi_items.VerticalLineROI):
            lineProjectionMode = 'Y'
            x = self.getPosition()
            roiStart = (x, 0)
            roiEnd = (x, 1)
        elif isinstance(self, roi_items.LineROI):
            lineProjectionMode = 'D'
            roiStart, roiEnd = self.getEndPoints()
        else:
            assert False

        return roiStart, roiEnd, lineProjectionMode

    def computeProfile(self, item):
        if not isinstance(item, items.ImageBase):
            raise TypeError("Unexpected class %s" % type(item))

        origin = item.getOrigin()
        scale = item.getScale()
        method = self.getProfileMethod()
        lineWidth = self.getProfileLineWidth()

        def createProfile2(currentData):
            coords, profile, _area, profileName, xLabel = core.createProfile(
                roiInfo=self._getRoiInfo(),
                currentData=currentData,
                origin=origin,
                scale=scale,
                lineWidth=lineWidth,
                method=method)
            return coords, profile, profileName, xLabel

        currentData = item.getValueData(copy=False)

        yLabel = "%s" % str(method).capitalize()
        coords, profile, title, xLabel = createProfile2(currentData)
        title = title + "; width = %d" % lineWidth

        # Use the axis names from the original plot
        profileManager = self.getProfileManager()
        plot = profileManager.getPlotWidget()
        title = _relabelAxes(plot, title)
        xLabel = _relabelAxes(plot, xLabel)

        if isinstance(item, items.ImageRgba):
            rgba = item.getData(copy=False)
            _coords, r, _profileName, _xLabel = createProfile2(rgba[..., 0])
            _coords, g, _profileName, _xLabel = createProfile2(rgba[..., 1])
            _coords, b, _profileName, _xLabel = createProfile2(rgba[..., 2])
            if rgba.shape[-1] == 4:
                _coords, a, _profileName, _xLabel = createProfile2(rgba[..., 3])
            else:
                a = [None]
            data = core.RgbaProfileData(
                coords=coords,
                profile=profile[0],
                profile_r=r[0],
                profile_g=g[0],
                profile_b=b[0],
                profile_a=a[0],
                title=title,
                xLabel=xLabel,
                yLabel=yLabel,
            )
        else:
            data = core.CurveProfileData(
                coords=coords,
                profile=profile[0],
                title=title,
                xLabel=xLabel,
                yLabel=yLabel,
            )
        return data


class ProfileImageHorizontalLineROI(roi_items.HorizontalLineROI,
                                    _DefaultImageProfileRoiMixIn):
    """ROI for an horizontal profile at a location of an image"""

    ICON = 'shape-horizontal'
    NAME = 'horizontal line profile'

    def __init__(self, parent=None):
        roi_items.HorizontalLineROI.__init__(self, parent=parent)
        _DefaultImageProfileRoiMixIn.__init__(self, parent=parent)


class ProfileImageVerticalLineROI(roi_items.VerticalLineROI,
                                  _DefaultImageProfileRoiMixIn):
    """ROI for a vertical profile at a location of an image"""

    ICON = 'shape-vertical'
    NAME = 'vertical line profile'

    def __init__(self, parent=None):
        roi_items.VerticalLineROI.__init__(self, parent=parent)
        _DefaultImageProfileRoiMixIn.__init__(self, parent=parent)


class ProfileImageLineROI(roi_items.LineROI,
                          _DefaultImageProfileRoiMixIn):
    """ROI for an image profile between 2 points.

    The X profile of this ROI is the projecting into one of the x/y axes,
    using its scale and its orientation.
    """

    ICON = 'shape-diagonal'
    NAME = 'line profile'

    def __init__(self, parent=None):
        roi_items.LineROI.__init__(self, parent=parent)
        _DefaultImageProfileRoiMixIn.__init__(self, parent=parent)


class ProfileImageDirectedLineROI(roi_items.LineROI,
                                  _DefaultImageProfileRoiMixIn):
    """ROI for an image profile between 2 points.

    The X profile of the line is displayed projected into the line itself,
    using its scale and its orientation. It's the distance from the origin.
    """

    ICON = 'shape-diagonal-directed'
    NAME = 'directed line profile'

    def __init__(self, parent=None):
        roi_items.LineROI.__init__(self, parent=parent)
        _DefaultImageProfileRoiMixIn.__init__(self, parent=parent)
        self._handleStart.setSymbol('o')

    def computeProfile(self, item):
        if not isinstance(item, items.ImageBase):
            raise TypeError("Unexpected class %s" % type(item))

        from silx.image.bilinear import BilinearImage

        origin = item.getOrigin()
        scale = item.getScale()
        method = self.getProfileMethod()
        lineWidth = self.getProfileLineWidth()
        currentData = item.getValueData(copy=False)

        roiInfo = self._getRoiInfo()
        roiStart, roiEnd, _lineProjectionMode = roiInfo

        startPt = ((roiStart[1] - origin[1]) / scale[1],
                   (roiStart[0] - origin[0]) / scale[0])
        endPt = ((roiEnd[1] - origin[1]) / scale[1],
                 (roiEnd[0] - origin[0]) / scale[0])

        if numpy.array_equal(startPt, endPt):
            return None

        bilinear = BilinearImage(currentData)
        profile = bilinear.profile_line(
            (startPt[0] - 0.5, startPt[1] - 0.5),
            (endPt[0] - 0.5, endPt[1] - 0.5),
            lineWidth,
            method=method)

        # Compute the line size
        lineSize = numpy.sqrt((roiEnd[1] - roiStart[1]) ** 2 +
                              (roiEnd[0] - roiStart[0]) ** 2)
        coords = numpy.linspace(0, lineSize, len(profile),
                                endpoint=True,
                                dtype=numpy.float32)

        title = _lineProfileTitle(*roiStart, *roiEnd)
        title = title + "; width = %d" % lineWidth
        xLabel = "√({xlabel}²+{ylabel}²)"
        yLabel = str(method).capitalize()

        # Use the axis names from the original plot
        profileManager = self.getProfileManager()
        plot = profileManager.getPlotWidget()
        xLabel = _relabelAxes(plot, xLabel)
        title = _relabelAxes(plot, title)

        data = core.CurveProfileData(
            coords=coords,
            profile=profile,
            title=title,
            xLabel=xLabel,
            yLabel=yLabel,
        )
        return data


class _ProfileCrossROI(roi_items.HandleBasedROI, core.ProfileRoiMixIn):

    """ROI to manage a cross of profiles

    It is managed using 2 sub ROIs for vertical and horizontal.
    """

    _kind = "Cross"
    """Label for this kind of ROI"""

    _plotShape = "point"
    """Plot shape which is used for the first interaction"""

    def __init__(self, parent=None):
        roi_items.HandleBasedROI.__init__(self, parent=parent)
        core.ProfileRoiMixIn.__init__(self, parent=parent)
        self.sigRegionChanged.connect(self.__regionChanged)
        self.sigAboutToBeRemoved.connect(self.__aboutToBeRemoved)
        self.__position = 0, 0
        self.__vline = None
        self.__hline = None
        self.__handle = self.addHandle()
        self.__handleLabel = self.addLabelHandle()
        self.__handleLabel.setText(self.getName())
        self.__inhibitReentance = utils.LockReentrant()
        self.computeProfile = None
        self.sigItemChanged.connect(self.__updateLineProperty)

        # Make sure the marker is over the ROIs
        self.__handle.setZValue(1)
        # Create the vline and the hline
        self._createSubRois()

    @docstring(roi_items.HandleBasedROI)
    def contains(self, position):
        roiPos = self.getPosition()
        return position[0] == roiPos[0] or position[1] == roiPos[1]

    def setFirstShapePoints(self, points):
        pos = points[0]
        self.setPosition(pos)

    def getPosition(self):
        """Returns the position of this ROI

        :rtype: numpy.ndarray
        """
        return self.__position

    def setPosition(self, pos):
        """Set the position of this ROI

        :param numpy.ndarray pos: 2d-coordinate of this point
        """
        self.__position = pos
        with utils.blockSignals(self.__handle):
            self.__handle.setPosition(*pos)
        with utils.blockSignals(self.__handleLabel):
            self.__handleLabel.setPosition(*pos)
        self.sigRegionChanged.emit()

    def handleDragUpdated(self, handle, origin, previous, current):
        if handle is self.__handle:
            self.setPosition(current)

    def __updateLineProperty(self, event=None, checkVisibility=True):
        if event == items.ItemChangedType.NAME:
            self.__handleLabel.setText(self.getName())
        elif event in [items.ItemChangedType.COLOR,
                     items.ItemChangedType.VISIBLE]:
            lines = []
            if self.__vline:
                lines.append(self.__vline)
            if self.__hline:
                lines.append(self.__hline)
            self._updateItemProperty(event, self, lines)

    def _createLines(self, parent):
        """Inherit this function to return 2 ROI objects for respectivly
        the horizontal, and the vertical lines."""
        raise NotImplementedError()

    def _setProfileManager(self, profileManager):
        core.ProfileRoiMixIn._setProfileManager(self, profileManager)
        # Connecting the vline and the hline
        roiManager = profileManager.getRoiManager()
        roiManager.addRoi(self.__vline)
        roiManager.addRoi(self.__hline)

    def _createSubRois(self):
        hline, vline = self._createLines(parent=None)
        for i, line in enumerate([vline, hline]):
            line.setPosition(self.__position[i])
            line.setEditable(True)
            line.setSelectable(True)
            line.setFocusProxy(self)
            line.setName("")
        self.__vline = vline
        self.__hline = hline
        vline.sigAboutToBeRemoved.connect(self.__vlineRemoved)
        vline.sigRegionChanged.connect(self.__vlineRegionChanged)
        hline.sigAboutToBeRemoved.connect(self.__hlineRemoved)
        hline.sigRegionChanged.connect(self.__hlineRegionChanged)

    def _getLines(self):
        return self.__hline, self.__vline

    def __regionChanged(self):
        if self.__inhibitReentance.locked():
            return
        x, y = self.getPosition()
        hline, vline = self._getLines()
        if hline is None:
            return
        with self.__inhibitReentance:
            hline.setPosition(y)
            vline.setPosition(x)

    def __vlineRegionChanged(self):
        if self.__inhibitReentance.locked():
            return
        pos = self.getPosition()
        vline = self.__vline
        pos = vline.getPosition(), pos[1]
        with self.__inhibitReentance:
            self.setPosition(pos)

    def __hlineRegionChanged(self):
        if self.__inhibitReentance.locked():
            return
        pos = self.getPosition()
        hline = self.__hline
        pos = pos[0], hline.getPosition()
        with self.__inhibitReentance:
            self.setPosition(pos)

    def __aboutToBeRemoved(self):
        vline = self.__vline
        hline = self.__hline
        # Avoid side remove signals
        if hline is not None:
            hline.sigAboutToBeRemoved.disconnect(self.__hlineRemoved)
            hline.sigRegionChanged.disconnect(self.__hlineRegionChanged)
        if vline is not None:
            vline.sigAboutToBeRemoved.disconnect(self.__vlineRemoved)
            vline.sigRegionChanged.disconnect(self.__vlineRegionChanged)
        # Clean up the child
        profileManager = self.getProfileManager()
        roiManager = profileManager.getRoiManager()
        if hline is not None:
            roiManager.removeRoi(hline)
            self.__hline = None
        if vline is not None:
            roiManager.removeRoi(vline)
            self.__vline = None

    def __hlineRemoved(self):
        self.__lineRemoved(isHline=True)

    def __vlineRemoved(self):
        self.__lineRemoved(isHline=False)

    def __lineRemoved(self, isHline):
        """If any of the lines is removed: disconnect this objects, and let the
        other one persist"""
        hline, vline = self._getLines()

        hline.sigAboutToBeRemoved.disconnect(self.__hlineRemoved)
        vline.sigAboutToBeRemoved.disconnect(self.__vlineRemoved)
        hline.sigRegionChanged.disconnect(self.__hlineRegionChanged)
        vline.sigRegionChanged.disconnect(self.__vlineRegionChanged)

        self.__hline = None
        self.__vline = None
        profileManager = self.getProfileManager()
        roiManager = profileManager.getRoiManager()
        if isHline:
            self.__releaseLine(vline)
        else:
            self.__releaseLine(hline)
        roiManager.removeRoi(self)

    def __releaseLine(self, line):
        """Release the line in order to make it independent"""
        line.setFocusProxy(None)
        line.setName(self.getName())
        line.setEditable(self.isEditable())
        line.setSelectable(self.isSelectable())


class ProfileImageCrossROI(_ProfileCrossROI):
    """ROI to manage a cross of profiles

    It is managed using 2 sub ROIs for vertical and horizontal.
    """

    ICON = 'shape-cross'
    NAME = 'cross profile'
    ITEM_KIND = items.ImageBase

    def _createLines(self, parent):
        vline = ProfileImageVerticalLineROI(parent=parent)
        hline = ProfileImageHorizontalLineROI(parent=parent)
        return hline, vline

    def setProfileMethod(self, method):
        """
        :param str method: method to compute the profile. Can be 'mean' or 'sum'
        """
        hline, vline = self._getLines()
        hline.setProfileMethod(method)
        vline.setProfileMethod(method)
        self.invalidateProperties()

    def getProfileMethod(self):
        hline, _vline = self._getLines()
        return hline.getProfileMethod()

    def setProfileLineWidth(self, width):
        hline, vline = self._getLines()
        hline.setProfileLineWidth(width)
        vline.setProfileLineWidth(width)
        self.invalidateProperties()

    def getProfileLineWidth(self):
        hline, _vline = self._getLines()
        return hline.getProfileLineWidth()


class _DefaultScatterProfileRoiMixIn(core.ProfileRoiMixIn):
    """Provide common behavior for silx default scatter profile ROI.
    """

    ITEM_KIND = items.Scatter

    def __init__(self, parent=None):
        core.ProfileRoiMixIn.__init__(self, parent=parent)
        self.__nPoints = 1024
        self.sigRegionChanged.connect(self.__regionChanged)

    def __regionChanged(self):
        self.invalidateProfile()

    # Number of points

    def getNPoints(self):
        """Returns the number of points of the profiles

        :rtype: int
        """
        return self.__nPoints

    def setNPoints(self, npoints):
        """Set the number of points of the profiles

        :param int npoints:
        """
        npoints = int(npoints)
        if npoints < 1:
            raise ValueError("Unsupported number of points: %d" % npoints)
        elif npoints != self.__nPoints:
            self.__nPoints = npoints
            self.invalidateProperties()
            self.invalidateProfile()

    def _computeProfile(self, scatter, x0, y0, x1, y1):
        """Compute corresponding profile

        :param float x0: Profile start point X coord
        :param float y0: Profile start point Y coord
        :param float x1: Profile end point X coord
        :param float y1: Profile end point Y coord
        :return: (points, values) profile data or None
        """
        future = scatter._getInterpolator()
        try:
            interpolator = future.result()
        except CancelledError:
            return None
        if interpolator is None:
            return None  # Cannot init an interpolator

        nPoints = self.getNPoints()
        points = numpy.transpose((
            numpy.linspace(x0, x1, nPoints, endpoint=True),
            numpy.linspace(y0, y1, nPoints, endpoint=True)))

        values = interpolator(points)

        if not numpy.any(numpy.isfinite(values)):
            return None  # Profile outside convex hull

        return points, values

    def computeProfile(self, item):
        """Update profile according to current ROI"""
        if not isinstance(item, items.Scatter):
            raise TypeError("Unexpected class %s" % type(item))

        # Get end points
        if isinstance(self, roi_items.LineROI):
            points = self.getEndPoints()
            x0, y0 = points[0]
            x1, y1 = points[1]
        elif isinstance(self, (roi_items.VerticalLineROI, roi_items.HorizontalLineROI)):
            profileManager = self.getProfileManager()
            plot = profileManager.getPlotWidget()

            if isinstance(self, roi_items.HorizontalLineROI):
                x0, x1 = plot.getXAxis().getLimits()
                y0 = y1 = self.getPosition()

            elif isinstance(self, roi_items.VerticalLineROI):
                x0 = x1 = self.getPosition()
                y0, y1 = plot.getYAxis().getLimits()
        else:
            raise RuntimeError('Unsupported ROI for profile: {}'.format(self.__class__))

        if x1 < x0 or (x1 == x0 and y1 < y0):
            # Invert points
            x0, y0, x1, y1 = x1, y1, x0, y0

        profile = self._computeProfile(item, x0, y0, x1, y1)
        if profile is None:
            return None

        title = _lineProfileTitle(x0, y0, x1, y1)
        points = profile[0]
        values = profile[1]

        if (numpy.abs(points[-1, 0] - points[0, 0]) >
                numpy.abs(points[-1, 1] - points[0, 1])):
            xProfile = points[:, 0]
            xLabel = '{xlabel}'
        else:
            xProfile = points[:, 1]
            xLabel = '{ylabel}'

        # Use the axis names from the original
        profileManager = self.getProfileManager()
        plot = profileManager.getPlotWidget()
        title = _relabelAxes(plot, title)
        xLabel = _relabelAxes(plot, xLabel)

        data = core.CurveProfileData(
            coords=xProfile,
            profile=values,
            title=title,
            xLabel=xLabel,
            yLabel='Profile',
        )
        return data


class ProfileScatterHorizontalLineROI(roi_items.HorizontalLineROI,
                                      _DefaultScatterProfileRoiMixIn):
    """ROI for an horizontal profile at a location of a scatter"""

    ICON = 'shape-horizontal'
    NAME = 'horizontal line profile'

    def __init__(self, parent=None):
        roi_items.HorizontalLineROI.__init__(self, parent=parent)
        _DefaultScatterProfileRoiMixIn.__init__(self, parent=parent)


class ProfileScatterVerticalLineROI(roi_items.VerticalLineROI,
                                    _DefaultScatterProfileRoiMixIn):
    """ROI for an horizontal profile at a location of a scatter"""

    ICON = 'shape-vertical'
    NAME = 'vertical line profile'

    def __init__(self, parent=None):
        roi_items.VerticalLineROI.__init__(self, parent=parent)
        _DefaultScatterProfileRoiMixIn.__init__(self, parent=parent)


class ProfileScatterLineROI(roi_items.LineROI,
                            _DefaultScatterProfileRoiMixIn):
    """ROI for an horizontal profile at a location of a scatter"""

    ICON = 'shape-diagonal'
    NAME = 'line profile'

    def __init__(self, parent=None):
        roi_items.LineROI.__init__(self, parent=parent)
        _DefaultScatterProfileRoiMixIn.__init__(self, parent=parent)


class ProfileScatterCrossROI(_ProfileCrossROI):
    """ROI to manage a cross of profiles for scatters.
    """

    ICON = 'shape-cross'
    NAME = 'cross profile'
    ITEM_KIND = items.Scatter

    def _createLines(self, parent):
        vline = ProfileScatterVerticalLineROI(parent=parent)
        hline = ProfileScatterHorizontalLineROI(parent=parent)
        return hline, vline

    def getNPoints(self):
        """Returns the number of points of the profiles

        :rtype: int
        """
        hline, _vline = self._getLines()
        return hline.getNPoints()

    def setNPoints(self, npoints):
        """Set the number of points of the profiles

        :param int npoints:
        """
        hline, vline = self._getLines()
        hline.setNPoints(npoints)
        vline.setNPoints(npoints)
        self.invalidateProperties()


class _DefaultScatterProfileSliceRoiMixIn(core.ProfileRoiMixIn):
    """Default ROI to allow to slice in the scatter data."""

    ITEM_KIND = items.Scatter

    def __init__(self, parent=None):
        core.ProfileRoiMixIn.__init__(self, parent=parent)
        self.__area = _SliceProfileArea(self)
        self.addItem(self.__area)
        self.sigRegionChanged.connect(self._regionChanged)
        self.sigPlotItemChanged.connect(self._updateArea)

    def _regionChanged(self):
        self.invalidateProfile()
        self._updateArea()

    def _updateArea(self):
        plotItem = self.getPlotItem()
        if plotItem is None:
            self.setLineStyle("-")
        else:
            self.setLineStyle("--")

    def _getSlice(self, item):
        position = self.getPosition()
        bounds = item.getCurrentVisualizationParameter(items.Scatter.VisualizationParameter.GRID_BOUNDS)
        if isinstance(self, roi_items.HorizontalLineROI):
            axis = 1
        elif isinstance(self, roi_items.VerticalLineROI):
            axis = 0
        else:
            assert False
        if bounds is None or position < bounds[0][axis] or position > bounds[1][axis]:
            # ROI outside of the scatter bound
            return None

        major_order = item.getCurrentVisualizationParameter(items.Scatter.VisualizationParameter.GRID_MAJOR_ORDER)
        assert major_order == 'row'
        max_grid_yy, max_grid_xx = item.getCurrentVisualizationParameter(items.Scatter.VisualizationParameter.GRID_SHAPE)

        xx, yy, _values, _xx_error, _yy_error = item.getData(copy=False)
        if isinstance(self, roi_items.HorizontalLineROI):
            axis = yy
            max_grid_first = max_grid_yy
            max_grid_second = max_grid_xx
            major_axis = major_order == 'column'
        elif isinstance(self, roi_items.VerticalLineROI):
            axis = xx
            max_grid_first = max_grid_xx
            max_grid_second = max_grid_yy
            major_axis = major_order == 'row'
        else:
            assert False

        def argnearest(array, value):
            array = numpy.abs(array - value)
            return numpy.argmin(array)

        if major_axis:
            # slice in the middle of the scatter
            actual_size_grid_second = len(axis) // max_grid_first
            start = actual_size_grid_second // 2 * max_grid_first
            vslice = axis[start:start + max_grid_first]
            if len(vslice) == 0:
                return None
            index = argnearest(vslice, position)
            slicing = slice(index, None, max_grid_first)
        else:
            # slice in the middle of the scatter
            actual_size_grid_second = len(axis) // max_grid_first
            vslice = axis[actual_size_grid_second // 2::max_grid_second]
            if len(vslice) == 0:
                return None
            index = argnearest(vslice, position)
            start = index * max_grid_second
            slicing = slice(start, start + max_grid_second)

        return slicing

    def computeProfile(self, item):
        if not isinstance(item, items.Scatter):
            raise TypeError("Unsupported %s item" % type(item))

        slicing = self._getSlice(item)
        if slicing is None:
            # ROI out of bounds
            return None

        _xx, _yy, values, _xx_error, _yy_error = item.getData(copy=False)
        profile = values[slicing]

        if isinstance(self, roi_items.HorizontalLineROI):
            title = "Horizontal slice"
            xLabel = "{xlabel} index"
        elif isinstance(self, roi_items.VerticalLineROI):
            title = "Vertical slice"
            xLabel = "{ylabel} index"
        else:
            assert False

        # Use the axis names from the original plot
        profileManager = self.getProfileManager()
        plot = profileManager.getPlotWidget()
        xLabel = _relabelAxes(plot, xLabel)

        data = core.CurveProfileData(
            coords=numpy.arange(len(profile)),
            profile=profile,
            title=title,
            xLabel=xLabel,
            yLabel="Profile",
        )
        return data


class ProfileScatterHorizontalSliceROI(roi_items.HorizontalLineROI,
                                       _DefaultScatterProfileSliceRoiMixIn):
    """ROI for an horizontal profile at a location of a scatter
    using data slicing.
    """

    ICON = 'slice-horizontal'
    NAME = 'horizontal data slice profile'

    def __init__(self, parent=None):
        roi_items.HorizontalLineROI.__init__(self, parent=parent)
        _DefaultScatterProfileSliceRoiMixIn.__init__(self, parent=parent)


class ProfileScatterVerticalSliceROI(roi_items.VerticalLineROI,
                                       _DefaultScatterProfileSliceRoiMixIn):
    """ROI for a vertical profile at a location of a scatter
    using data slicing.
    """

    ICON = 'slice-vertical'
    NAME = 'vertical data slice profile'

    def __init__(self, parent=None):
        roi_items.VerticalLineROI.__init__(self, parent=parent)
        _DefaultScatterProfileSliceRoiMixIn.__init__(self, parent=parent)


class ProfileScatterCrossSliceROI(_ProfileCrossROI):
    """ROI to manage a cross of slicing profiles on scatters.
    """

    ICON = 'slice-cross'
    NAME = 'cross data slice profile'
    ITEM_KIND = items.Scatter

    def _createLines(self, parent):
        vline = ProfileScatterVerticalSliceROI(parent=parent)
        hline = ProfileScatterHorizontalSliceROI(parent=parent)
        return hline, vline


class _DefaultImageStackProfileRoiMixIn(_DefaultImageProfileRoiMixIn):

    ITEM_KIND = items.ImageStack

    def __init__(self, parent=None):
        super(_DefaultImageStackProfileRoiMixIn, self).__init__(parent=parent)
        self.__profileType = "1D"
        """Kind of profile"""

    def getProfileType(self):
        return self.__profileType

    def setProfileType(self, kind):
        assert kind in ["1D", "2D"]
        if self.__profileType == kind:
            return
        self.__profileType = kind
        self.invalidateProperties()
        self.invalidateProfile()

    def computeProfile(self, item):
        if not isinstance(item, items.ImageStack):
            raise TypeError("Unexpected class %s" % type(item))

        kind = self.getProfileType()
        if kind == "1D":
            result = _DefaultImageProfileRoiMixIn.computeProfile(self, item)
            # z = item.getStackPosition()
            return result

        assert kind == "2D"

        def createProfile2(currentData):
            coords, profile, _area, profileName, xLabel = core.createProfile(
                roiInfo=self._getRoiInfo(),
                currentData=currentData,
                origin=origin,
                scale=scale,
                lineWidth=self.getProfileLineWidth(),
                method=method)
            return coords, profile, profileName, xLabel

        currentData = numpy.array(item.getStackData(copy=False))
        origin = item.getOrigin()
        scale = item.getScale()
        colormap = item.getColormap()
        method = self.getProfileMethod()

        coords, profile, profileName, xLabel = createProfile2(currentData)

        profileManager = self.getProfileManager()
        plot = profileManager.getPlotWidget()

        data = core.ImageProfileData(
            coords=coords,
            profile=profile,
            title=_relabelAxes(plot, profileName),
            xLabel=_relabelAxes(plot, xLabel),
            yLabel="Profile",
            colormap=colormap,
        )
        return data


class ProfileImageStackHorizontalLineROI(roi_items.HorizontalLineROI,
                                         _DefaultImageStackProfileRoiMixIn):
    """ROI for an horizontal profile at a location of a stack of images"""

    ICON = 'shape-horizontal'
    NAME = 'horizontal line profile'

    def __init__(self, parent=None):
        roi_items.HorizontalLineROI.__init__(self, parent=parent)
        _DefaultImageStackProfileRoiMixIn.__init__(self, parent=parent)


class ProfileImageStackVerticalLineROI(roi_items.VerticalLineROI,
                                       _DefaultImageStackProfileRoiMixIn):
    """ROI for an vertical profile at a location of a stack of images"""

    ICON = 'shape-vertical'
    NAME = 'vertical line profile'

    def __init__(self, parent=None):
        roi_items.VerticalLineROI.__init__(self, parent=parent)
        _DefaultImageStackProfileRoiMixIn.__init__(self, parent=parent)


class ProfileImageStackLineROI(roi_items.LineROI,
                               _DefaultImageStackProfileRoiMixIn):
    """ROI for an vertical profile at a location of a stack of images"""

    ICON = 'shape-diagonal'
    NAME = 'line profile'

    def __init__(self, parent=None):
        roi_items.LineROI.__init__(self, parent=parent)
        _DefaultImageStackProfileRoiMixIn.__init__(self, parent=parent)


class ProfileImageStackCrossROI(ProfileImageCrossROI):
    """ROI for an vertical profile at a location of a stack of images"""

    ICON = 'shape-cross'
    NAME = 'cross profile'
    ITEM_KIND = items.ImageStack

    def _createLines(self, parent):
        vline = ProfileImageStackVerticalLineROI(parent=parent)
        hline = ProfileImageStackHorizontalLineROI(parent=parent)
        return hline, vline

    def getProfileType(self):
        hline, _vline = self._getLines()
        return hline.getProfileType()

    def setProfileType(self, kind):
        hline, vline = self._getLines()
        hline.setProfileType(kind)
        vline.setProfileType(kind)
        self.invalidateProperties()



import pytest
import logging
import os


logger = logging.getLogger(__name__)


def _set_qt_binding(binding):
    if binding is not None:
        binding = binding.lower()
        if binding == "pyqt5":
            logger.info("Force using PyQt5")
            import PyQt5.QtCore  # noqa
        elif binding == "pyside2":
            logger.info("Force using PySide2")
            import PySide2.QtCore  # noqa
        elif binding == "pyside6":
            logger.info("Force using PySide6")
            import PySide6.QtCore  # noqa
        elif binding == "pyqt6":
            logger.info("Force using PyQt6")
            import PyQt6.QtCore  # noqa
        else:
            raise ValueError("Qt binding '%s' is unknown" % binding)


def pytest_addoption(parser):
    parser.addoption("--qt-binding", type=str, default=None, dest="qt_binding",
                     help="Force using a Qt binding: 'PyQt5', 'PySide2', 'PySide6', 'PyQt6'")
    parser.addoption("--no-gui", dest="gui", default=True,
                     action="store_false",
                     help="Disable the test of the graphical use interface")
    parser.addoption("--no-opengl", dest="opengl", default=True,
                     action="store_false",
                     help="Disable tests using OpenGL")
    parser.addoption("--no-opencl", dest="opencl", default=True,
                     action="store_false",
                     help="Disable the test of the OpenCL part")
    parser.addoption("--low-mem", dest="low_mem", default=False,
                     action="store_true",
                     help="Disable test with large memory consumption (>100Mbyte")


def pytest_configure(config):
    if not config.getoption('opencl', True):
        os.environ['SILX_OPENCL'] = 'False'  # Disable OpenCL support in silx

    _set_qt_binding(config.option.qt_binding)


@pytest.fixture(scope="session")
def test_options(request):
    from .test import utils
    options = utils._TestOptions()
    options.configure(request.config.option)
    yield options


@pytest.fixture(scope="class")
def test_options_class_attr(request, test_options):
    """Provides test_options as class attribute

    Used as transition from TestCase to pytest
    """
    request.cls.test_options = test_options


@pytest.fixture(scope="session")
def use_opengl(test_options):
    """Fixture to flag test using a OpenGL.

    This can be skipped with `--no-opengl`.
    """
    if not test_options.WITH_GL_TEST:
        pytest.skip(test_options.WITH_GL_TEST_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def use_opencl(test_options):
    """Fixture to flag test using a OpenCL.

    This can be skipped with `--no-opencl`.
    """
    if not test_options.WITH_OPENCL_TEST:
        pytest.skip(test_options.WITH_OPENCL_TEST_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def use_large_memory(test_options):
    """Fixture to flag test using a large memory consumption.

    This can be skipped with `--low-mem`.
    """
    if test_options.TEST_LOW_MEM:
        pytest.skip(test_options.TEST_LOW_MEM_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def use_gui(test_options):
    """Fixture to flag test using GUI.

    This can be skipped with `--no-gui`.
    """
    if not test_options.WITH_QT_TEST:
        pytest.skip(test_options.WITH_QT_TEST_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def qapp(use_gui, xvfb, request):
    _set_qt_binding(request.config.option.qt_binding)

    from silx.gui import qt
    app = qt.QApplication.instance()
    if app is None:
        app = qt.QApplication([])
    try:
        yield app
    finally:
        if app is not None:
            app.closeAllWindows()


@pytest.fixture
def qapp_utils(qapp):
    """Helper containing method to deal with QApplication and widget"""
    from silx.gui.utils.testutils import TestCaseQt
    utils = TestCaseQt()
    utils.setUpClass()
    utils.setUp()
    yield utils
    utils.tearDown()
    utils.tearDownClass()



# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Weakref utils for compatibility between Python 2 and Python 3 or for
extended features.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "15/09/2016"


import weakref
import types
import inspect


def ref(object, callback=None):
    """Returns a weak reference to object. The original object can be retrieved
    by calling the reference object if the referent is still alive. If the
    referent is no longer alive, calling the reference object will cause None
    to be returned.

    The signature is the same as the standard `weakref` library, but it returns
    `WeakMethod` if the object is a bound method.

    :param object: An object
    :param func callback: If provided, and the returned weakref object is
        still alive, the callback will be called when the object is about to
        be finalized. The weak reference object will be passed as the only
        parameter to the callback. Then the referent will no longer be
        available.
    :return: A weak reference to the object
    """
    if inspect.ismethod(object):
        return WeakMethod(object, callback)
    else:
        return weakref.ref(object, callback)


def proxy(object, callback=None):
    """Return a proxy to object which uses a weak reference. This supports use
    of the proxy in most contexts instead of requiring the explicit
    dereferencing used with weak reference objects.

    The signature is the same as the standard `weakref` library, but it returns
    `WeakMethodProxy` if the object is a bound method.

    :param object: An object
    :param func callback: If provided, and the returned weakref object is
        still alive, the callback will be called when the object is about to
        be finalized. The weak reference object will be passed as the only
        parameter to the callback. Then the referent will no longer be
        available.
    :return: A proxy to a weak reference of the object
    """
    if inspect.ismethod(object):
        return WeakMethodProxy(object, callback)
    else:
        return weakref.proxy(object, callback)


class WeakMethod(object):
    """Wraps a callable object like a function or a bound method.
    Feature callback when the object is about to be finalized.
    Provids the same interface as a normal weak reference.
    """

    def __init__(self, function, callback=None):
        """
        Constructor
        :param function: Function/method to be called
        :param callback: If callback is provided and not None,
            and the returned weakref object is still alive, the
            callback will be called when the object is about to
            be finalized; the weak reference object will be passed
            as the only parameter to the callback; the referent will
            no longer be available
        """
        self.__callback = callback

        if inspect.ismethod(function):
            # it is a bound method
            self.__obj = weakref.ref(function.__self__, self.__call_callback)
            self.__method = weakref.ref(function.__func__, self.__call_callback)
        else:
            self.__obj = None
            self.__method = weakref.ref(function, self.__call_callback)

    def __call_callback(self, ref):
        """Called when the object is about to be finalized"""
        if not self.is_alive():
            return
        self.__obj = None
        self.__method = None
        if self.__callback is not None:
            self.__callback(self)

    def __call__(self):
        """Return a callable function or None if the WeakMethod is dead."""
        if self.__obj is not None:
            method = self.__method()
            obj = self.__obj()
            if method is None or obj is None:
                return None
            return types.MethodType(method, obj)
        elif self.__method is not None:
            return self.__method()
        else:
            return None

    def is_alive(self):
        """True if the WeakMethod is still alive"""
        return self.__method is not None

    def __eq__(self, other):
        """Check it another obect is equal to this.

        :param object other: Object to compare with
        """
        if isinstance(other, WeakMethod):
            if not self.is_alive():
                return False
            return self.__obj == other.__obj and self.__method == other.__method
        return False

    def __ne__(self, other):
        """Check it another obect is not equal to this.

        :param object other: Object to compare with
        """
        if isinstance(other, WeakMethod):
            if not self.is_alive():
                return False
            return self.__obj != other.__obj or self.__method != other.__method
        return True

    def __hash__(self):
        """Returns the hash for the object."""
        return self.__obj.__hash__() ^ self.__method.__hash__()


class WeakMethodProxy(WeakMethod):
    """Wraps a callable object like a function or a bound method
    with a weakref proxy.
    """
    def __call__(self, *args, **kwargs):
        """Dereference the method and call it if the method is still alive.
        Else raises an ReferenceError.

        :raises: ReferenceError, if the method is not alive
        """
        fn = super(WeakMethodProxy, self).__call__()
        if fn is None:
            raise ReferenceError("weakly-referenced object no longer exists")
        return fn(*args, **kwargs)


class WeakList(list):
    """Manage a list of weaked references.
    When an object is dead, the list is flaged as invalid.
    If expected the list is cleaned up to remove dead objects.
    """

    def __init__(self, enumerator=()):
        """Create a WeakList

        :param iterator enumerator: A list of object to initialize the
            list
        """
        list.__init__(self)
        self.__list = []
        self.__is_valid = True
        for obj in enumerator:
            self.append(obj)

    def __invalidate(self, ref):
        """Flag the list as invalidated. The list contains dead references."""
        self.__is_valid = False

    def __create_ref(self, obj):
        """Create a weakref from an object. It uses the `ref` module function.
        """
        return ref(obj, self.__invalidate)

    def __clean(self):
        """Clean the list from dead references"""
        if self.__is_valid:
            return
        self.__list = [ref for ref in self.__list if ref() is not None]
        self.__is_valid = True

    def __iter__(self):
        """Iterate over objects of the list"""
        for ref in self.__list:
            obj = ref()
            if obj is not None:
                yield obj

    def __len__(self):
        """Count item on the list"""
        self.__clean()
        return len(self.__list)

    def __getitem__(self, key):
        """Returns the object at the requested index

        :param key: Indexes to get
        :type key: int or slice
        """
        self.__clean()
        data = self.__list[key]
        if isinstance(data, list):
            result = [ref() for ref in data]
        else:
            result = data()
        return result

    def __setitem__(self, key, obj):
        """Set an item at an index

        :param key: Indexes to set
        :type key: int or slice
        """
        self.__clean()
        if isinstance(key, slice):
            objs = [self.__create_ref(o) for o in obj]
            self.__list[key] = objs
        else:
            obj_ref = self.__create_ref(obj)
            self.__list[key] = obj_ref

    def __delitem__(self, key):
        """Delete an Indexes item of this list

        :param key: Index to delete
        :type key: int or slice
         """
        self.__clean()
        del self.__list[key]

    def __delslice__(self, i, j):
        """Looks to be used in Python 2.7"""
        self.__delitem__(slice(i, j, None))

    def __setslice__(self, i, j, sequence):
        """Looks to be used in Python 2.7"""
        self.__setitem__(slice(i, j, None), sequence)

    def __getslice__(self, i, j):
        """Looks to be used in Python 2.7"""
        return self.__getitem__(slice(i, j, None))

    def __reversed__(self):
        """Returns a copy of the reverted list"""
        reversed_list = reversed(list(self))
        return WeakList(reversed_list)

    def __contains__(self, obj):
        """Returns true if the object is in the list"""
        ref = self.__create_ref(obj)
        return ref in self.__list

    def __add__(self, other):
        """Returns a WeakList containing this list an the other"""
        l = WeakList(self)
        l.extend(other)
        return l

    def __iadd__(self, other):
        """Add objects to this list inplace"""
        self.extend(other)
        return self

    def __mul__(self, n):
        """Returns a WeakList containing n-duplication object of this list"""
        return WeakList(list(self) * n)

    def __imul__(self, n):
        """N-duplication of the objects to this list inplace"""
        self.__list *= n
        return self

    def append(self, obj):
        """Add an object at the end of the list"""
        ref = self.__create_ref(obj)
        self.__list.append(ref)

    def count(self, obj):
        """Returns the number of occurencies of an object"""
        ref = self.__create_ref(obj)
        return self.__list.count(ref)

    def extend(self, other):
        """Append the list with all objects from another list"""
        for obj in other:
            self.append(obj)

    def index(self, obj):
        """Returns the index of an object"""
        ref = self.__create_ref(obj)
        return self.__list.index(ref)

    def insert(self, index, obj):
        """Insert an object at the requested index"""
        ref = self.__create_ref(obj)
        self.__list.insert(index, ref)

    def pop(self, index=-1):
        """Remove and return an object at the requested index"""
        self.__clean()
        obj = self.__list.pop(index)()
        return obj

    def remove(self, obj):
        """Remove an object from the list"""
        ref = self.__create_ref(obj)
        self.__list.remove(ref)

    def reverse(self):
        """Reverse the list inplace"""
        self.__list.reverse()

    def sort(self, key=None, reverse=False):
        """Sort the list inplace.
        Not very efficient.
        """
        sorted_list = list(self)
        sorted_list.sort(key=key, reverse=reverse)
        self.__list = []
        self.extend(sorted_list)

    def __str__(self):
        unref_list = list(self)
        return "WeakList(%s)" % str(unref_list)

    def __repr__(self):
        unref_list = list(self)
        return "WeakList(%s)" % repr(unref_list)



# /*##########################################################################
#
# Copyright (c) 2020-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides a base class for PlotWidget OpenGL backend primitives
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "02/07/2020"


class RenderContext:
    """Context with which to perform OpenGL rendering.

    :param numpy.ndarray matrix: 4x4 transform matrix to use for rendering
    :param bool isXLog: Whether X axis is log scale or not
    :param bool isYLog: Whether Y axis is log scale or not
    :param float dpi: Number of device pixels per inch
    """

    def __init__(self, matrix=None, isXLog=False, isYLog=False, dpi=96., plotFrame=None):
        self.matrix = matrix
        """Current transformation matrix"""

        self.__isXLog = isXLog
        self.__isYLog = isYLog
        self.__dpi = dpi
        self.__plotFrame = plotFrame

    @property
    def isXLog(self):
        """True if X axis is using log scale"""
        return self.__isXLog

    @property
    def isYLog(self):
        """True if Y axis is using log scale"""
        return self.__isYLog

    @property
    def dpi(self):
        """Number of device pixels per inch"""
        return self.__dpi

    @property
    def plotFrame(self):
        """Current PlotFrame"""
        return self.__plotFrame


class GLPlotItem:
    """Base class for primitives used in the PlotWidget OpenGL backend"""

    def __init__(self):
        self.yaxis = 'left'
        "YAxis this item is attached to (either 'left' or 'right')"

    def pick(self, x, y):
        """Perform picking at given position.

        :param float x: X coordinate in plot data frame of reference
        :param float y: Y coordinate in plot data frame of reference
        :returns:
           Result of picking as a list of indices or None if nothing picked
        :rtype: Union[List[int],None]
        """
        return None

    def render(self, context):
        """Performs OpenGL rendering of the item.

        :param RenderContext context: Rendering context information
        """
        pass

    def discard(self):
        """Discards OpenGL resources this item has created."""
        pass

    def isInitialized(self) -> bool:
        """Returns True if resources where initialized and requires `discard`.
        """
        return True



#!/usr/bin/env python3
# /*##########################################################################
#
# Copyright (c) 2015-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Run the tests of the project.

This script expects a suite function in <project_package>.test,
which returns a unittest.TestSuite.

Test coverage dependencies: coverage, lxml.
"""

__authors__ = ["Jérôme Kieffer", "Thomas Vincent"]
__date__ = "30/09/2020"
__license__ = "MIT"

import importlib
import logging
import os
import subprocess
import sys
import sysconfig


# Capture all default warnings
logging.captureWarnings(True)
import warnings
warnings.simplefilter('default')

logger = logging.getLogger("run_tests")
logger.setLevel(logging.WARNING)

logger.info("Python %s %s", sys.version, tuple.__itemsize__ * 8)


try:
    import numpy
except Exception as error:
    logger.warning("Numpy missing: %s", error)
else:
    logger.info("Numpy %s", numpy.version.version)

try:
    import h5py
except Exception as error:
    logger.warning("h5py missing: %s", error)
else:
    logger.info("h5py %s", h5py.version.version)


def get_project_name(root_dir):
    """Retrieve project name by running python setup.py --name in root_dir.

    :param str root_dir: Directory where to run the command.
    :return: The name of the project stored in root_dir
    """
    logger.debug("Getting project name in %s", root_dir)
    p = subprocess.Popen([sys.executable, "setup.py", "--name"],
                         shell=False, cwd=root_dir, stdout=subprocess.PIPE)
    name, _stderr_data = p.communicate()
    logger.debug("subprocess ended with rc= %s", p.returncode)
    return name.split()[-1].decode('ascii')


def is_debug_python():
    """Returns true if the Python interpreter is in debug mode."""
    if sysconfig.get_config_var("Py_DEBUG"):
        return True

    return hasattr(sys, "gettotalrefcount")


def build_project(name, root_dir):
    """Run python setup.py build for the project.

    Build directory can be modified by environment variables.

    :param str name: Name of the project.
    :param str root_dir: Root directory of the project
    :return: The path to the directory were build was performed
    """
    platform = sysconfig.get_platform()
    architecture = "lib.%s-%i.%i" % (platform,
                                     sys.version_info[0], sys.version_info[1])
    if is_debug_python():
        architecture += "-pydebug"

    if os.environ.get("PYBUILD_NAME") == name:
        # we are in the debian packaging way
        home = os.environ.get("PYTHONPATH", "").split(os.pathsep)[-1]
    elif os.environ.get("BUILDPYTHONPATH"):
        home = os.path.abspath(os.environ.get("BUILDPYTHONPATH", ""))
    else:
        home = os.path.join(root_dir, "build", architecture)

    logger.warning("Building %s to %s", name, home)
    cmd = [sys.executable, "setup.py", "build"]
    if home:
        cmd += ["--build-lib", home]
    p = subprocess.Popen(cmd, shell=False, cwd=root_dir)
    logger.debug("subprocess ended with rc= %s", p.wait())

    if os.path.isdir(home):
        return home
    alt_home = os.path.join(os.path.dirname(home), "lib")
    if os.path.isdir(alt_home):
        return alt_home


def import_project_module(project_name, project_dir):
    """Import project module, from the system of from the project directory"""
    if "--installed" in sys.argv:
        try:
            module = importlib.import_module(project_name)
        except Exception:
            logger.error("Cannot run tests on installed version: %s not installed or raising error.",
                         project_name)
            raise
    else:  # Use built source
        build_dir = build_project(project_name, project_dir)
        if build_dir is None:
            logging.error("Built project is not available !!! investigate")
        sys.path.insert(0, build_dir)
        logger.warning("Patched sys.path, added: '%s'", build_dir)
        module = importlib.import_module(project_name)
    return module


if __name__ == "__main__":  # Needed for multiprocessing support on Windows
    import pytest

    PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
    PROJECT_NAME = get_project_name(PROJECT_DIR)
    logger.info("Project name: %s", PROJECT_NAME)

    project_module = import_project_module(PROJECT_NAME, PROJECT_DIR)
    PROJECT_VERSION = getattr(project_module, 'version', '')
    PROJECT_PATH = project_module.__path__[0]

    def normalize_option(option):
        option_parts = option.split(os.path.sep)
        if option_parts == ["src", "silx"]:
            return PROJECT_PATH
        if option_parts[:2] == ["src", "silx"]:
            return os.path.join(PROJECT_PATH, *option_parts[2:])
        return option

    args = [normalize_option(p) for p in sys.argv[1:] if p != "--installed"]

    # Run test on PROJECT_PATH if nothing is specified
    without_options = [a for a in args if not a.startswith("-")]
    if len(without_options) == 0:
        args += [PROJECT_PATH]

    argv = ["--rootdir", PROJECT_PATH] + args
    sys.exit(pytest.main(argv))



# /*##########################################################################
#
# Copyright (c) 2014-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module implements date-time labels layout on graph axes."""

__authors__ = ["P. Kenter"]
__license__ = "MIT"
__date__ = "04/04/2018"


import datetime as dt
import enum
import logging
import math
import time

import dateutil.tz

from dateutil.relativedelta import relativedelta

from .ticklayout import niceNumGeneric

_logger = logging.getLogger(__name__)


MICROSECONDS_PER_SECOND = 1000000
SECONDS_PER_MINUTE = 60
SECONDS_PER_HOUR = 60 * SECONDS_PER_MINUTE
SECONDS_PER_DAY = 24 * SECONDS_PER_HOUR
SECONDS_PER_YEAR = 365.25 * SECONDS_PER_DAY
SECONDS_PER_MONTH_AVERAGE = SECONDS_PER_YEAR / 12 # Seconds per average month


# No dt.timezone in Python 2.7 so we use dateutil.tz.tzutc
_EPOCH = dt.datetime(1970, 1, 1, tzinfo=dateutil.tz.tzutc())

def timestamp(dtObj):
    """ Returns POSIX timestamp of a datetime objects.

    If the dtObj object has a timestamp() method (python 3.3), this is
    used. Otherwise (e.g. python 2.7) it is calculated here.

    The POSIX timestamp is a floating point value of the number of seconds
    since the start of an epoch (typically 1970-01-01). For details see:
    https://docs.python.org/3/library/datetime.html#datetime.datetime.timestamp

    :param datetime.datetime dtObj: date-time representation.
    :return: POSIX timestamp
    :rtype: float
    """
    if hasattr(dtObj, "timestamp"):
        return dtObj.timestamp()
    else:
        # Back ported from Python 3.5
        if dtObj.tzinfo is None:
            return time.mktime((dtObj.year, dtObj.month, dtObj.day,
                                dtObj.hour, dtObj.minute, dtObj.second,
                                -1, -1, -1)) + dtObj.microsecond / 1e6
        else:
            return (dtObj - _EPOCH).total_seconds()


@enum.unique
class DtUnit(enum.Enum):
    YEARS = 0
    MONTHS = 1
    DAYS = 2
    HOURS = 3
    MINUTES = 4
    SECONDS = 5
    MICRO_SECONDS = 6  # a fraction of a second


def getDateElement(dateTime, unit):
    """ Picks the date element with the unit from the dateTime

    E.g. getDateElement(datetime(1970, 5, 6), DtUnit.Day) will return 6

    :param datetime dateTime: date/time to pick from
    :param DtUnit unit: The unit describing the date element.
    """
    if unit == DtUnit.YEARS:
        return dateTime.year
    elif unit == DtUnit.MONTHS:
        return dateTime.month
    elif unit == DtUnit.DAYS:
        return dateTime.day
    elif unit == DtUnit.HOURS:
        return dateTime.hour
    elif unit == DtUnit.MINUTES:
        return dateTime.minute
    elif unit == DtUnit.SECONDS:
        return dateTime.second
    elif unit == DtUnit.MICRO_SECONDS:
        return dateTime.microsecond
    else:
        raise ValueError("Unexpected DtUnit: {}".format(unit))


def setDateElement(dateTime, value, unit):
    """ Returns a copy of dateTime with the tickStep unit set to value

    :param datetime.datetime: date time object
    :param int value: value to set
    :param DtUnit unit: unit
    :return: datetime.datetime
    """
    intValue = int(value)
    _logger.debug("setDateElement({}, {} (int={}), {})"
                  .format(dateTime, value, intValue, unit))

    year = dateTime.year
    month = dateTime.month
    day = dateTime.day
    hour = dateTime.hour
    minute = dateTime.minute
    second = dateTime.second
    microsecond = dateTime.microsecond

    if unit == DtUnit.YEARS:
        year = intValue
    elif unit == DtUnit.MONTHS:
        month = intValue
    elif unit == DtUnit.DAYS:
        day = intValue
    elif unit == DtUnit.HOURS:
        hour = intValue
    elif unit == DtUnit.MINUTES:
        minute = intValue
    elif unit == DtUnit.SECONDS:
        second = intValue
    elif unit == DtUnit.MICRO_SECONDS:
        microsecond = intValue
    else:
        raise ValueError("Unexpected DtUnit: {}".format(unit))

    _logger.debug("creating date time {}"
        .format((year, month, day, hour, minute, second, microsecond)))

    return dt.datetime(year, month, day, hour, minute, second, microsecond,
                       tzinfo=dateTime.tzinfo)



def roundToElement(dateTime, unit):
    """ Returns a copy of dateTime rounded to given unit

    :param datetime.datetime: date time object
    :param DtUnit unit: unit
    :return: datetime.datetime
    """
    year = dateTime.year
    month = dateTime.month
    day = dateTime.day
    hour = dateTime.hour
    minute = dateTime.minute
    second = dateTime.second
    microsecond = dateTime.microsecond

    if unit.value < DtUnit.YEARS.value:
        pass # Never round years
    if unit.value < DtUnit.MONTHS.value:
        month = 1
    if unit.value < DtUnit.DAYS.value:
        day = 1
    if unit.value < DtUnit.HOURS.value:
        hour = 0
    if unit.value < DtUnit.MINUTES.value:
        minute = 0
    if unit.value < DtUnit.SECONDS.value:
        second = 0
    if unit.value < DtUnit.MICRO_SECONDS.value:
        microsecond = 0

    result = dt.datetime(year, month, day, hour, minute, second, microsecond,
                         tzinfo=dateTime.tzinfo)

    return result


def addValueToDate(dateTime, value, unit):
    """ Adds a value with unit to a dateTime.

    Uses dateutil.relativedelta.relativedelta from the standard library to do
    the actual math. This function doesn't allow for fractional month or years,
    so month and year are truncated to integers before adding.

    :param datetime dateTime: date time
    :param float value: value to be added
    :param DtUnit unit: of the value
    :return:
    :raises ValueError: unit is unsupported or result is out of datetime bounds
    """
    #logger.debug("addValueToDate({}, {}, {})".format(dateTime, value, unit))

    if unit == DtUnit.YEARS:
        intValue = int(value) # floats not implemented in relativeDelta(years)
        return dateTime + relativedelta(years=intValue)
    elif unit == DtUnit.MONTHS:
        intValue = int(value) # floats not implemented in relativeDelta(mohths)
        return dateTime + relativedelta(months=intValue)
    elif unit == DtUnit.DAYS:
        return dateTime + relativedelta(days=value)
    elif unit == DtUnit.HOURS:
        return dateTime + relativedelta(hours=value)
    elif unit == DtUnit.MINUTES:
        return dateTime + relativedelta(minutes=value)
    elif unit == DtUnit.SECONDS:
        return dateTime + relativedelta(seconds=value)
    elif unit == DtUnit.MICRO_SECONDS:
        return dateTime + relativedelta(microseconds=value)
    else:
        raise ValueError("Unexpected DtUnit: {}".format(unit))


def bestUnit(durationInSeconds):
    """ Gets the best tick spacing given a duration in seconds.

    :param durationInSeconds: time span duration in seconds
    :return: DtUnit enumeration.
    """

    # Based on; https://stackoverflow.com/a/2144398/
    # If the duration is longer than two years the tick spacing will be in
    # years. Else, if the duration is longer than two months, the spacing will
    # be in months, Etcetera.
    #
    # This factor differs per unit. As a baseline it is 2, but for instance,
    # for Months this needs to be higher (3>), This because it is impossible to
    # have partial months so the tick spacing is always at least 1 month. A
    # duration of two months would result in two ticks, which is too few.
    # months would then results

    if durationInSeconds > SECONDS_PER_YEAR * 3:
        return (durationInSeconds / SECONDS_PER_YEAR, DtUnit.YEARS)
    elif durationInSeconds > SECONDS_PER_MONTH_AVERAGE * 3:
        return (durationInSeconds / SECONDS_PER_MONTH_AVERAGE, DtUnit.MONTHS)
    elif durationInSeconds > SECONDS_PER_DAY * 2:
        return (durationInSeconds / SECONDS_PER_DAY, DtUnit.DAYS)
    elif durationInSeconds > SECONDS_PER_HOUR * 2:
        return (durationInSeconds / SECONDS_PER_HOUR, DtUnit.HOURS)
    elif durationInSeconds > SECONDS_PER_MINUTE * 2:
        return (durationInSeconds / SECONDS_PER_MINUTE, DtUnit.MINUTES)
    elif durationInSeconds > 1 * 2:
        return (durationInSeconds, DtUnit.SECONDS)
    else:
        return (durationInSeconds * MICROSECONDS_PER_SECOND,
                DtUnit.MICRO_SECONDS)


NICE_DATE_VALUES = {
    DtUnit.YEARS: [1, 2, 5, 10],
    DtUnit.MONTHS: [1, 2, 3, 4, 6, 12],
    DtUnit.DAYS: [1, 2, 3, 7, 14, 28],
    DtUnit.HOURS: [1, 2, 3, 4, 6, 12],
    DtUnit.MINUTES: [1, 2, 3, 5, 10, 15, 30],
    DtUnit.SECONDS: [1, 2, 3, 5, 10, 15, 30],
    DtUnit.MICRO_SECONDS : [1.0, 2.0, 5.0, 10.0], # floats for microsec
}


def bestFormatString(spacing, unit):
    """ Finds the best format string given the spacing and DtUnit.

    If the spacing is a fractional number < 1 the format string will take this
    into account

    :param spacing: spacing between ticks
    :param DtUnit unit:
    :return: Format string for use in strftime
    :rtype: str
    """
    isSmall = spacing < 1

    if unit == DtUnit.YEARS:
        return "%Y-m" if isSmall else "%Y"
    elif unit == DtUnit.MONTHS:
        return "%Y-%m-%d" if isSmall else "%Y-%m"
    elif unit == DtUnit.DAYS:
        return "%H:%M" if isSmall else "%Y-%m-%d"
    elif unit == DtUnit.HOURS:
        return "%H:%M" if isSmall else "%H:%M"
    elif unit == DtUnit.MINUTES:
        return "%H:%M:%S" if isSmall else "%H:%M"
    elif unit == DtUnit.SECONDS:
        return "%S.%f" if isSmall else "%H:%M:%S"
    elif unit == DtUnit.MICRO_SECONDS:
        return "%S.%f"
    else:
        raise ValueError("Unexpected DtUnit: {}".format(unit))


def niceDateTimeElement(value, unit, isRound=False):
    """ Uses the Nice Numbers algorithm to determine a nice value.

    The fractions are optimized for the unit of the date element.
    """

    niceValues = NICE_DATE_VALUES[unit]
    elemValue = niceNumGeneric(value, niceValues, isRound=isRound)

    if unit == DtUnit.YEARS or unit == DtUnit.MONTHS:
        elemValue = max(1, int(elemValue))

    return elemValue


def findStartDate(dMin, dMax, nTicks):
    """ Rounds a date down to the nearest nice number of ticks
    """
    assert dMax >= dMin, \
        "dMin ({}) should come before dMax ({})".format(dMin, dMax)

    if dMin == dMax:
        # Fallback when range is smaller than microsecond resolution
        return dMin, 1, DtUnit.MICRO_SECONDS

    delta = dMax - dMin
    lengthSec = delta.total_seconds()
    _logger.debug("findStartDate: {}, {} (duration = {} sec, {} days)"
                  .format(dMin, dMax, lengthSec, lengthSec / SECONDS_PER_DAY))

    length, unit = bestUnit(lengthSec)
    niceLength = niceDateTimeElement(length, unit)

    _logger.debug("Length: {:8.3f} {} (nice = {})"
                  .format(length, unit.name, niceLength))

    niceSpacing = niceDateTimeElement(niceLength / nTicks, unit, isRound=True)

    _logger.debug("Spacing: {:8.3f} {} (nice = {})"
                  .format(niceLength / nTicks, unit.name, niceSpacing))

    dVal = getDateElement(dMin, unit)

    if unit == DtUnit.MONTHS: # TODO: better rounding?
        niceVal = math.floor((dVal-1) / niceSpacing) * niceSpacing + 1
    elif unit == DtUnit.DAYS:
        niceVal = math.floor((dVal-1) / niceSpacing) * niceSpacing + 1
    else:
        niceVal = math.floor(dVal / niceSpacing) * niceSpacing

    if unit == DtUnit.YEARS and niceVal <= dt.MINYEAR:
        niceVal = max(1, niceSpacing)

    _logger.debug("StartValue: dVal = {}, niceVal: {} ({})"
                  .format(dVal, niceVal, unit.name))

    startDate = roundToElement(dMin, unit)
    startDate = setDateElement(startDate, niceVal, unit)

    return startDate, niceSpacing, unit


def dateRange(dMin, dMax, step, unit, includeFirstBeyond = False):
    """ Generates a range of dates

    :param datetime dMin: start date
    :param datetime dMax: end date
    :param int step: the step size
    :param DtUnit unit: the unit of the step size
    :param bool includeFirstBeyond: if True the first date later than dMax will
        be included in the range. If False (the default), the last generated
        datetime will always be smaller than dMax.
    :return:
    """
    if (unit == DtUnit.YEARS or unit == DtUnit.MONTHS or
        unit == DtUnit.MICRO_SECONDS):
        # No support for fractional month or year and resolution is microsecond
        # In those cases, make sure the step is at least 1
        step = max(1, step)
    else:
        assert step > 0, "tickstep is 0"

    dateTime = dMin
    while dateTime < dMax:
        yield dateTime
        try:
            dateTime = addValueToDate(dateTime, step, unit)
        except ValueError:
            return  # current dateTime is out of datetime bounds

    if includeFirstBeyond:
        yield dateTime



def calcTicks(dMin, dMax, nTicks):
    """Returns tick positions.

    :param datetime.datetime dMin: The min value on the axis
    :param datetime.datetime dMax: The max value on the axis
    :param int nTicks: The target number of ticks. The actual number of found
        ticks may differ.
    :returns: (list of datetimes, DtUnit) tuple
    """
    _logger.debug("Calc calcTicks({}, {}, nTicks={})"
                  .format(dMin, dMax, nTicks))

    startDate, niceSpacing, unit = findStartDate(dMin, dMax, nTicks)

    result = []
    for d in dateRange(startDate, dMax, niceSpacing, unit,
                       includeFirstBeyond=True):
        result.append(d)

    return result, niceSpacing, unit


def calcTicksAdaptive(dMin, dMax, axisLength, tickDensity):
    """ Calls calcTicks with a variable number of ticks, depending on axisLength
    """
    # At least 2 ticks
    nticks = max(2, int(round(tickDensity * axisLength)))
    return  calcTicks(dMin, dMax, nticks)








# /*##########################################################################
# Copyright (C) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests of the calibration module"""

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "14/05/2018"


import unittest

import numpy

from silx.math.calibration import NoCalibration, LinearCalibration, \
    ArrayCalibration, FunctionCalibration


X = numpy.array([3.14, 2.73, 1337])


class TestNoCalibration(unittest.TestCase):
    def setUp(self):
        self.calib = NoCalibration()

    def testIsAffine(self):
        self.assertTrue(self.calib.is_affine())

    def testSlope(self):
        self.assertEqual(self.calib.get_slope(), 1.)

    def testYIntercept(self):
        self.assertEqual(self.calib(0.),
                         0.)

    def testCall(self):
        self.assertTrue(numpy.array_equal(self.calib(X), X))


class TestLinearCalibration(unittest.TestCase):
    def setUp(self):
        self.y_intercept = 1.5
        self.slope = 2.5
        self.calib = LinearCalibration(y_intercept=self.y_intercept,
                                       slope=self.slope)

    def testIsAffine(self):
        self.assertTrue(self.calib.is_affine())

    def testSlope(self):
        self.assertEqual(self.calib.get_slope(), self.slope)

    def testYIntercept(self):
        self.assertEqual(self.calib(0.),
                         self.y_intercept)

    def testCall(self):
        self.assertTrue(numpy.array_equal(self.calib(X),
                                          self.y_intercept + self.slope * X))


class TestArrayCalibration(unittest.TestCase):
    def setUp(self):
        self.arr = numpy.array([45.2, 25.3, 666., -8.])
        self.calib = ArrayCalibration(self.arr)
        self.affine_calib = ArrayCalibration([0.1, 0.2, 0.3])

    def testIsAffine(self):
        self.assertFalse(self.calib.is_affine())
        self.assertTrue(self.affine_calib.is_affine())

    def testSlope(self):
        with self.assertRaises(AttributeError):
            self.calib.get_slope()
        self.assertEqual(self.affine_calib.get_slope(),
                         0.1)

    def testYIntercept(self):
        self.assertEqual(self.calib(0),
                         self.arr[0])

    def testCall(self):
        with self.assertRaises(ValueError):
            # X is an array with a different shape
            self.calib(X)

        with self.assertRaises(ValueError):
            # floats are not valid indices
            self.calib(3.14)

        self.assertTrue(
            numpy.array_equal(self.calib([1, 2, 3, 4]),
                              self.arr))

        for idx, value in enumerate(self.arr):
            self.assertEqual(self.calib(idx), value)


class TestFunctionCalibration(unittest.TestCase):
    def setUp(self):
        self.non_affine_fun = numpy.sin
        self.non_affine_calib = FunctionCalibration(self.non_affine_fun)

        self.affine_fun = lambda x: 52. * x + 0.01
        self.affine_calib = FunctionCalibration(self.affine_fun,
                                                is_affine=True)

    def testIsAffine(self):
        self.assertFalse(self.non_affine_calib.is_affine())
        self.assertTrue(self.affine_calib.is_affine())

    def testSlope(self):
        with self.assertRaises(AttributeError):
            self.non_affine_calib.get_slope()
        self.assertAlmostEqual(self.affine_calib.get_slope(),
                               52.)

    def testCall(self):
        for x in X:
            self.assertAlmostEqual(self.non_affine_calib(x),
                                   self.non_affine_fun(x))
            self.assertAlmostEqual(self.affine_calib(x),
                                   self.affine_fun(x))



# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
histogramnd benchmarks, vs numpy.histogramdd (bin counts and weights).
"""

import numpy as np

import time

from silx.math import histogramnd


def print_times(t0s, t1s, t2s, t3s):
    c_times = t1s - t0s
    np_times = t2s - t1s
    np_w_times = t3s - t2s

    time_txt = 'min : {0: <7.3f}; max : {1: <7.3f}; avg : {2: <7.3f}'

    print('\tTimes :')
    print('\tC     : ' + time_txt.format(c_times.min(),
                                         c_times.max(),
                                         c_times.mean()))
    print('\tNP    : ' + time_txt.format(np_times.min(),
                                         np_times.max(),
                                         np_times.mean()))
    print('\tNP(W) : ' + time_txt.format(np_w_times.min(),
                                         np_w_times.max(),
                                         np_w_times.mean()))


def commpare_results(txt,
                     times,
                     result_c,
                     result_np,
                     result_np_w,
                     sample,
                     weights,
                     raise_ex=False):

    if result_np:
        hits_cmp = np.array_equal(result_c[0], result_np[0])
    else:
        hits_cmp = None

    if result_np_w and result_c[1] is not None:
        weights_cmp = np.array_equal(result_c[1], result_np_w[0])
    else:
        weights_cmp = None

    if((hits_cmp is not None and not hits_cmp) or
       (weights_cmp is not None and not weights_cmp)):
        err_txt = (txt + ' : results arent the same : '
                   'hits : {0}, '
                   'weights : {1}.'
                   ''.format('OK' if hits_cmp else 'NOK',
                             'OK' if weights_cmp else 'NOK'))
        print('\t' + err_txt)
        if raise_ex:
            raise ValueError(err_txt)
        return False

    result_txt = ' : results OK. c : {0: <7.3f};'.format(times[0])
    if result_np or result_np_w:
        result_txt += (' np : {0: <7.3f}; '
                       'np (weights) {1: <7.3f}.'
                       ''.format(times[1], times[2]))
    print('\t' + txt + result_txt)
    return True


def benchmark(n_loops,
              sample_shape,
              sample_rng,
              weights_rng,
              histo_range,
              n_bins,
              weight_min,
              weight_max,
              last_bin_closed,
              dtype=np.double,
              do_weights=True,
              do_numpy=True):

    int_min = 0
    int_max = 100000

    sample = np.random.randint(int_min,
                               high=int_max,
                               size=sample_shape).astype(np.double)
    sample = (sample_rng[0] +
              (sample - int_min) *
              (sample_rng[1] - sample_rng[0]) /
              (int_max - int_min))
    sample = sample.astype(dtype)

    if do_weights:
        weights = np.random.randint(int_min,
                                    high=int_max,
                                    size=(ssetup.pyample_shape[0],))
        weights = weights.astype(np.double)
        weights = (weights_rng[0] +
                   (weights - int_min) *
                   (weights_rng[1] - weights_rng[0]) /
                   (int_max - int_min))
    else:
        weights = None

    t0s = []
    t1s = []
    t2s = []
    t3s = []

    for i in range(n_loops):
        t0s.append(time.time())
        result_c = histogramnd(sample,
                               histo_range,
                               n_bins,
                               weights=weights,
                               weight_min=weight_min,
                               weight_max=weight_max,
                               last_bin_closed=last_bin_closed)
        t1s.append(time.time())
        if do_numpy:
            result_np = np.histogramdd(sample,
                                       bins=n_bins,
                                       range=histo_range)
            t2s.append(time.time())
            result_np_w = np.histogramdd(sample,
                                         bins=n_bins,
                                         range=histo_range,
                                         weights=weights)
            t3s.append(time.time())
        else:
            result_np = None
            result_np_w = None
            t2s.append(0)
            t3s.append(0)

        commpare_results('Run {0}'.format(i),
                         [t1s[-1] - t0s[-1], t2s[-1] - t1s[-1], t3s[-1] - t2s[-1]],
                         result_c,
                         result_np,
                         result_np_w,
                         sample,
                         weights)

    print_times(np.array(t0s), np.array(t1s), np.array(t2s), np.array(t3s))


def run_benchmark(dtype=np.double,
                  do_weights=True,
                  do_numpy=True):
    n_loops = 5

    weights_rng = [0., 100.]
    sample_rng = [0., 100.]

    weight_min = None
    weight_max = None
    last_bin_closed = True

    # ====================================================
    # ====================================================
    # 1D
    # ====================================================
    # ====================================================

    print('==========================')
    print(' 1D [{0}]'.format(dtype))
    print('==========================')
    sample_shape = (10**7,)
    histo_range = [[0., 100.]]
    n_bins = 30

    benchmark(n_loops,
              sample_shape,
              sample_rng,
              weights_rng,
              histo_range,
              n_bins,
              weight_min,
              weight_max,
              last_bin_closed,
              dtype=dtype,
              do_weights=True,
              do_numpy=do_numpy)

    # ====================================================
    # ====================================================
    # 2D
    # ====================================================
    # ====================================================

    print('==========================')
    print(' 2D [{0}]'.format(dtype))
    print('==========================')
    sample_shape = (10**7, 2)
    histo_range = [[0., 100.], [0., 100.]]
    n_bins = 30

    benchmark(n_loops,
              sample_shape,
              sample_rng,
              weights_rng,
              histo_range,
              n_bins,
              weight_min,
              weight_max,
              last_bin_closed,
              dtype=dtype,
              do_weights=True,
              do_numpy=do_numpy)

    # ====================================================
    # ====================================================
    # 3D
    # ====================================================
    # ====================================================

    print('==========================')
    print(' 3D [{0}]'.format(dtype))
    print('==========================')
    sample_shape = (10**7, 3)
    histo_range = np.array([[0., 100.], [0., 100.], [0., 100.]])
    n_bins = 30

    benchmark(n_loops,
              sample_shape,
              sample_rng,
              weights_rng,
              histo_range,
              n_bins,
              weight_min,
              weight_max,
              last_bin_closed,
              dtype=dtype,
              do_weights=True,
              do_numpy=do_numpy)

if __name__ == '__main__':
    types = (np.double, np.int32, np.float32,)

    for t in types:
        run_benchmark(t,
                      do_weights=True,
                      do_numpy=True)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Example demonstrating the use of the widget CompareImages
"""

import sys
import logging
import numpy
import argparse
import os

import silx.io
from silx.gui import qt
import silx.test.utils
from silx.io.url import DataUrl
from silx.gui.plot.CompareImages import CompareImages
from silx.gui.widgets.UrlSelectionTable import UrlSelectionTable

_logger = logging.getLogger(__name__)

import fabio

try:
    import PIL
except ImportError:
    _logger.debug("Backtrace", exc_info=True)
    PIL = None


class CompareImagesSelection(qt.QMainWindow):
    def __init__(self, backend):
        qt.QMainWindow.__init__(self, parent=None)
        self._plot = CompareImages(parent=self, backend=backend)

        self._selectionTable = UrlSelectionTable(parent=self)
        self._dockWidgetMenu = qt.QDockWidget(parent=self)
        self._dockWidgetMenu.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetMenu.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._dockWidgetMenu.setWidget(self._selectionTable)
        self.addDockWidget(qt.Qt.LeftDockWidgetArea, self._dockWidgetMenu)

        self.setCentralWidget(self._plot)

        self._selectionTable.sigImageAChanged.connect(self._updateImageA)
        self._selectionTable.sigImageBChanged.connect(self._updateImageB)

    def setUrls(self, urls):
        for url in urls:
            self._selectionTable.addUrl(url)

    def setFiles(self, files):
        urls = list()
        for _file in files:
            if os.path.isfile(_file):
                urls.append(DataUrl(file_path=_file, scheme=None))
        urls.sort(key=lambda url: url.path())
        window.setUrls(urls)
        window._selectionTable.setSelection(url_img_a=urls[0].path(),
                                            url_img_b=urls[1].path())

    def clear(self):
        self._plot.clear()
        self._selectionTable.clear()

    def _updateImageA(self, urlpath):
        self._updateImage(urlpath, self._plot.setImage1)

    def _updateImage(self, urlpath, fctptr):
        def getData():
            _url = silx.io.url.DataUrl(path=urlpath)
            for scheme in ('silx', 'fabio'):
                try:
                    dataImg = silx.io.utils.get_data(
                        silx.io.url.DataUrl(file_path=_url.file_path(),
                                            data_slice=_url.data_slice(),
                                            data_path=_url.data_path(),
                                            scheme=scheme))
                except:
                    _logger.debug("Error while loading image with %s" % scheme,
                                  exc_info=True)
                else:
                    # TODO: check is an image
                    return dataImg
            return None

        data = getData()
        if data is not None:
            fctptr(data)

    def _updateImageB(self, urlpath):
        self._updateImage(urlpath, self._plot.setImage2)


def createTestData():
    data = numpy.arange(100 * 100)
    data = (data % 100) / 5.0
    data = numpy.sin(data)
    data1 = data.copy()
    data1.shape = 100, 100
    data2 = silx.test.utils.add_gaussian_noise(data, stdev=0.1)
    data2.shape = 100, 100
    return data1, data2


def loadImage(filename):
    try:
        return silx.io.get_data(filename)
    except Exception:
        _logger.debug("Error while loading image with silx.io", exc_info=True)

    try:
        return fabio.open(filename).data
    except Exception:
        _logger.debug("Error while loading image with fabio", exc_info=True)

    if PIL is not None:
        try:
            return numpy.asarray(PIL.Image.open(filename))
        except Exception:
            _logger.debug("Error while loading image with PIL", exc_info=True)

    raise Exception("Impossible to load '%s' with the available image libraries" % filename)


file_description = """
Image data to compare (HDF5 file with path, EDF files, JPEG/PNG image files).
Data from HDF5 files can be accessed using dataset path and slicing as an URL: silx:../my_file.h5?path=/entry/data&slice=10
EDF file frames also can can be accessed using URL: fabio:../my_file.edf?slice=10
Using URL in command like usually have to be quoted: "URL".
"""


def createParser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'files',
        nargs=argparse.ZERO_OR_MORE,
        help=file_description)
    parser.add_argument(
        '--debug',
        dest="debug",
        action="store_true",
        default=False,
        help='Set logging system in debug mode')
    parser.add_argument(
        '--testdata',
        dest="testdata",
        action="store_true",
        default=False,
        help='Use synthetic images to test the application')
    parser.add_argument(
        '--use-opengl-plot',
        dest="use_opengl_plot",
        action="store_true",
        default=False,
        help='Use OpenGL for plots (instead of matplotlib)')
    return parser


if __name__ == "__main__":
    parser = createParser()
    options = parser.parse_args(sys.argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    if options.use_opengl_plot:
        backend = "gl"
    else:
        backend = "mpl"

    app = qt.QApplication([])
    if options.testdata or len(options.files) == 2:
        if options.testdata:
            _logger.info("Generate test data")
            data1, data2 = createTestData()
        else:
            data1 = loadImage(options.files[0])
            data2 = loadImage(options.files[1])
        window = CompareImages(backend=backend)
        window.setData(data1, data2)
    else:
        data = options.files
        window = CompareImagesSelection(backend=backend)
        window.setFiles(options.files)

    window.setVisible(True)
    app.exec()



# /*##########################################################################
#
# Copyright (c) 2015-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["P. Kenter"]
__license__ = "MIT"
__date__ = "06/04/2018"


import datetime as dt
import pytest


from silx.gui.plot._utils.dtime_ticklayout import calcTicks, DtUnit, SECONDS_PER_YEAR


def testSmallMonthlySpacing():
    """Tests a range that did result in a spacing of less than 1 month.
    It is impossible to add fractional month so the unit must be in days
    """
    from dateutil import parser

    d1 = parser.parse("2017-01-03 13:15:06.000044")
    d2 = parser.parse("2017-03-08 09:16:16.307584")
    _ticks, _units, spacing = calcTicks(d1, d2, nTicks=4)

    assert spacing == DtUnit.DAYS


def testNoCrash():
    """Creates many combinations of and number-of-ticks and end-dates;
    tests that it doesn't give an exception and returns a reasonable number
    of ticks.
    """
    d1 = dt.datetime(2017, 1, 3, 13, 15, 6, 44)

    value = 100e-6  # Start at 100 micro sec range.

    while value <= 200 * SECONDS_PER_YEAR:

        d2 = d1 + dt.timedelta(microseconds=value * 1e6)  # end date range

        for numTicks in range(2, 12):
            ticks, _, _ = calcTicks(d1, d2, numTicks)

            margin = 2.5
            assert (
                numTicks / margin <= len(ticks) <= numTicks * margin
            ), "Condition {} <= {} <= {} failed for # ticks={} and d2={}:".format(
                numTicks / margin, len(ticks), numTicks * margin, numTicks, d2
            )

        value = value * 1.5  # let date period grow exponentially


@pytest.mark.parametrize(
    "dMin, dMax",
    [
        (dt.datetime(1, 1, 1), dt.datetime(400, 1, 1)),
        (dt.datetime(4000, 1, 1), dt.datetime(9999, 1, 1)),
        (dt.datetime(1, 1, 1), dt.datetime(9999, 12, 23)),
    ],
)
def testCalcTicksOutOfBoundTicks(dMin, dMax):
    """Test tick generation with values leading to out-of-bound ticks"""
    ticks, _, unit = calcTicks(dMin, dMax, nTicks=5)
    assert len(ticks) != 0
    assert unit == DtUnit.YEARS



# /*##########################################################################
#
# Copyright (c) 2004-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""WaitingPushButton module
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "26/04/2017"

from .. import qt
from .. import icons


class WaitingPushButton(qt.QPushButton):
    """Button which allows to display a waiting status when, for example,
    something is still computing.

    The component is graphically disabled when it is in waiting. Then we
    overwrite the enabled method to dissociate the 2 concepts:
    graphically enabled/disabled, and enabled/disabled

    .. image:: img/WaitingPushButton.png
    """

    def __init__(self, parent=None, text=None, icon=None):
        """Constructor

        :param str text: Text displayed on the button
        :param qt.QIcon icon: Icon displayed on the button
        :param qt.QWidget parent: Parent of the widget
        """
        if icon is not None:
            qt.QPushButton.__init__(self, icon, text, parent)
        elif text is not None:
            qt.QPushButton.__init__(self, text, parent)
        else:
            qt.QPushButton.__init__(self, parent)

        self.__waiting = False
        self.__enabled = True
        self.__icon = icon
        self.__disabled_when_waiting = True
        self.__waitingIcon = icons.getWaitIcon()

    def sizeHint(self):
        """Returns the recommended size for the widget.

        This implementation of the recommended size always consider there is an
        icon. In this way it avoid to update the layout when the waiting icon
        is displayed.
        """
        self.ensurePolished()

        w = 0
        h = 0

        opt = qt.QStyleOptionButton()
        self.initStyleOption(opt)

        # Content with icon
        # no condition, assume that there is an icon to avoid blinking
        # when the widget switch to waiting state
        ih = opt.iconSize.height()
        iw = opt.iconSize.width() + 4
        w += iw
        h = max(h, ih)

        # Content with text
        text = self.text()
        isEmpty = text == ""
        if isEmpty:
            text = "XXXX"
        fm = self.fontMetrics()
        textSize = fm.size(qt.Qt.TextShowMnemonic, text)
        if not isEmpty or w == 0:
            w += textSize.width()
        if not isEmpty or h == 0:
            h = max(h, textSize.height())

        # Content with menu indicator
        opt.rect.setSize(qt.QSize(w, h))  # PM_MenuButtonIndicator depends on the height
        if self.menu() is not None:
            w += self.style().pixelMetric(qt.QStyle.PM_MenuButtonIndicator, opt, self)

        contentSize = qt.QSize(w, h)
        sizeHint = self.style().sizeFromContents(qt.QStyle.CT_PushButton, opt, contentSize, self)
        if qt.BINDING in ('PySide2', 'PyQt5'):  # Qt6: globalStrut not available
            sizeHint = sizeHint.expandedTo(qt.QApplication.globalStrut())
        return sizeHint

    def setDisabledWhenWaiting(self, isDisabled):
        """Enable or disable the auto disable behaviour when the button is waiting.

        :param bool isDisabled: Enable the auto-disable behaviour
        """
        if self.__disabled_when_waiting == isDisabled:
            return
        self.__disabled_when_waiting = isDisabled
        self.__updateVisibleEnabled()

    def isDisabledWhenWaiting(self):
        """Returns true if the button is auto disabled when it is waiting.

        :rtype: bool
        """
        return self.__disabled_when_waiting

    disabledWhenWaiting = qt.Property(bool, isDisabledWhenWaiting, setDisabledWhenWaiting)
    """Property to enable/disable the auto disabled state when the button is waiting."""

    def __setWaitingIcon(self, icon):
        """Called when the waiting icon is updated. It is called every frames
        of the animation.

        :param qt.QIcon icon: The new waiting icon
        """
        qt.QPushButton.setIcon(self, icon)

    def setIcon(self, icon):
        """Set the button icon. If the button is waiting, the icon is not
        visible directly, but will be visible when the waiting state will be
        removed.

        :param qt.QIcon icon: An icon
        """
        self.__icon = icon
        self.__updateVisibleIcon()

    def getIcon(self):
        """Returns the icon set to the button. If the widget is waiting
        it is not returning the visible icon, but the one requested by
        the application (the one displayed when the widget is not in
        waiting state).

        :rtype: qt.QIcon
        """
        return self.__icon

    icon = qt.Property(qt.QIcon, getIcon, setIcon)
    """Property providing access to the icon."""

    def __updateVisibleIcon(self):
        """Update the visible icon according to the state of the widget."""
        if not self.isWaiting():
            icon = self.__icon
        else:
            icon = self.__waitingIcon.currentIcon()
        if icon is None:
            icon = qt.QIcon()
        qt.QPushButton.setIcon(self, icon)

    def setEnabled(self, enabled):
        """Set the enabled state of the widget.

        :param bool enabled: The enabled state
        """
        if self.__enabled == enabled:
            return
        self.__enabled = enabled
        self.__updateVisibleEnabled()

    def isEnabled(self):
        """Returns the enabled state of the widget.

        :rtype: bool
        """
        return self.__enabled

    enabled = qt.Property(bool, isEnabled, setEnabled)
    """Property providing access to the enabled state of the widget"""

    def __updateVisibleEnabled(self):
        """Update the visible enabled state according to the state of the
        widget."""
        if self.__disabled_when_waiting:
            enabled = not self.isWaiting() and self.__enabled
        else:
            enabled = self.__enabled
        qt.QPushButton.setEnabled(self, enabled)

    def setWaiting(self, waiting):
        """Set the waiting state of the widget.

        :param bool waiting: Requested state"""
        if self.__waiting == waiting:
            return
        self.__waiting = waiting

        if self.__waiting:
            self.__waitingIcon.register(self)
            self.__waitingIcon.iconChanged.connect(self.__setWaitingIcon)
        else:
            # unregister only if the object is registred
            self.__waitingIcon.unregister(self)
            self.__waitingIcon.iconChanged.disconnect(self.__setWaitingIcon)

        self.__updateVisibleEnabled()
        self.__updateVisibleIcon()

    def isWaiting(self):
        """Returns true if the widget is in waiting state.

        :rtype: bool"""
        return self.__waiting

    @qt.Slot()
    def wait(self):
        """Enable the waiting state."""
        self.setWaiting(True)

    @qt.Slot()
    def stopWaiting(self):
        """Disable the waiting state."""
        self.setWaiting(False)

    @qt.Slot()
    def swapWaiting(self):
        """Swap the waiting state."""
        self.setWaiting(not self.isWaiting())



# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test for silx.gui.hdf5 module"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "17/01/2018"


import unittest
import time
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.gui.utils.testutils import SignalListener
from silx.gui.widgets.ThreadPoolPushButton import ThreadPoolPushButton
from silx.utils.testutils import LoggingValidator


class TestThreadPoolPushButton(TestCaseQt):

    def setUp(self):
        super(TestThreadPoolPushButton, self).setUp()
        self._result = []

    def waitForPendingOperations(self, object):
        for i in range(50):
            if not object.hasPendingOperations():
                break
            self.qWait(10)
        else:
            raise RuntimeError("Still waiting for a pending operation")

    def _trace(self, name, delay=0):
        self._result.append(name)
        if delay != 0:
            time.sleep(delay / 1000.0)

    def _compute(self):
        return "result"

    def _computeFail(self):
        raise Exception("exception")

    def testExecute(self):
        button = ThreadPoolPushButton()
        button.setCallable(self._trace, "a", 0)
        button.executeCallable()
        time.sleep(0.1)
        self.assertListEqual(self._result, ["a"])
        self.waitForPendingOperations(button)

    def testMultiExecution(self):
        button = ThreadPoolPushButton()
        button.setCallable(self._trace, "a", 0)
        number = qt.silxGlobalThreadPool().maxThreadCount()
        for _ in range(number):
            button.executeCallable()
        self.waitForPendingOperations(button)
        self.assertListEqual(self._result, ["a"] * number)

    def testSaturateThreadPool(self):
        button = ThreadPoolPushButton()
        button.setCallable(self._trace, "a", 100)
        number = qt.silxGlobalThreadPool().maxThreadCount() * 2
        for _ in range(number):
            button.executeCallable()
        self.waitForPendingOperations(button)
        self.assertListEqual(self._result, ["a"] * number)

    def testSuccess(self):
        listener = SignalListener()
        button = ThreadPoolPushButton()
        button.setCallable(self._compute)
        button.beforeExecuting.connect(listener.partial(test="be"))
        button.started.connect(listener.partial(test="s"))
        button.succeeded.connect(listener.partial(test="result"))
        button.failed.connect(listener.partial(test="Unexpected exception"))
        button.finished.connect(listener.partial(test="f"))
        button.executeCallable()
        self.qapp.processEvents()
        time.sleep(0.1)
        self.qapp.processEvents()
        result = listener.karguments(argumentName="test")
        self.assertListEqual(result, ["be", "s", "result", "f"])

    def testFail(self):
        listener = SignalListener()
        button = ThreadPoolPushButton()
        button.setCallable(self._computeFail)
        button.beforeExecuting.connect(listener.partial(test="be"))
        button.started.connect(listener.partial(test="s"))
        button.succeeded.connect(listener.partial(test="Unexpected success"))
        button.failed.connect(listener.partial(test="exception"))
        button.finished.connect(listener.partial(test="f"))
        with LoggingValidator('silx.gui.widgets.ThreadPoolPushButton', error=1):
            button.executeCallable()
            self.qapp.processEvents()
            time.sleep(0.1)
            self.qapp.processEvents()
        result = listener.karguments(argumentName="test")
        self.assertListEqual(result, ["be", "s", "exception", "f"])
        listener.clear()



#!/usr/bin/env python
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2013-2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""
This module provides a class for CBF byte offset compression/decompression.
"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "11/10/2018"
__status__ = "production"


import functools
import os
import numpy
from ..common import ocl, pyopencl
from ..processing import BufferDescription, EventDescription, OpenclProcessing

import logging
logger = logging.getLogger(__name__)

if pyopencl:
    import pyopencl.version
    if pyopencl.version.VERSION < (2016, 0):
        from pyopencl.scan import GenericScanKernel, GenericDebugScanKernel
    else:
        from pyopencl.algorithm import GenericScanKernel
        from pyopencl.scan import GenericDebugScanKernel
else:
    logger.warning("No PyOpenCL, no byte-offset, please see fabio")


class ByteOffset(OpenclProcessing):
    """Perform the byte offset compression/decompression on the GPU

        See :class:`OpenclProcessing` for optional arguments description.

        :param int raw_size:
            Size of the raw stream for decompression.
            It can be (slightly) larger than the array.
        :param int dec_size:
            Size of the decompression output array
            (mandatory for decompression)
        """

    def __init__(self, raw_size=None, dec_size=None,
                 ctx=None, devicetype="all",
                 platformid=None, deviceid=None,
                 block_size=None, profile=False):
        OpenclProcessing.__init__(self, ctx=ctx, devicetype=devicetype,
                                  platformid=platformid, deviceid=deviceid,
                                  block_size=block_size, profile=profile)
        if self.block_size is None:
            self.block_size = self.device.max_work_group_size
        wg = self.block_size

        buffers = [BufferDescription("counter", 1, numpy.int32, None)]

        if raw_size is None:
            self.raw_size = -1
            self.padded_raw_size = -1
        else:
            self.raw_size = int(raw_size)
            self.padded_raw_size = int((self.raw_size + wg - 1) & ~(wg - 1))
            buffers += [
                BufferDescription("raw", self.padded_raw_size, numpy.int8, None),
                BufferDescription("mask", self.padded_raw_size, numpy.int32, None),
                BufferDescription("values", self.padded_raw_size, numpy.int32, None),
                BufferDescription("exceptions", self.padded_raw_size, numpy.int32, None)
            ]

        if dec_size is None:
            self.dec_size = None
        else:
            self.dec_size = numpy.int32(dec_size)
            buffers += [
                BufferDescription("data_float", self.dec_size, numpy.float32, None),
                BufferDescription("data_int", self.dec_size, numpy.int32, None)
            ]

        self.allocate_buffers(buffers, use_array=True)

        self.compile_kernels([os.path.join("codec", "byte_offset")])
        self.kernels.__setattr__("scan", self._init_double_scan())
        self.kernels.__setattr__("compression_scan",
                                 self._init_compression_scan())

    def _init_double_scan(self):
        """"generates a double scan on indexes and values in one operation"""
        arguments = "__global int *value", "__global int *index"
        int2 = pyopencl.tools.get_or_register_dtype("int2")
        input_expr = "index[i]>0 ? (int2)(0, 0) : (int2)(value[i], 1)"
        scan_expr = "a+b"
        neutral = "(int2)(0,0)"
        output_statement = "value[i] = item.s0; index[i+1] = item.s1;"

        if self.block_size > 256:
            knl = GenericScanKernel(self.ctx,
                                    dtype=int2,
                                    arguments=arguments,
                                    input_expr=input_expr,
                                    scan_expr=scan_expr,
                                    neutral=neutral,
                                    output_statement=output_statement)
        else:  # MacOS on CPU
            knl = GenericDebugScanKernel(self.ctx,
                                         dtype=int2,
                                         arguments=arguments,
                                         input_expr=input_expr,
                                         scan_expr=scan_expr,
                                         neutral=neutral,
                                         output_statement=output_statement)
        return knl

    def decode(self, raw, as_float=False, out=None):
        """This function actually performs the decompression by calling the kernels

        :param numpy.ndarray raw: The compressed data as a 1D numpy array of char.
        :param bool as_float: True to decompress as float32,
                              False (default) to decompress as int32
        :param pyopencl.array out: pyopencl array in which to place the result.
        :return: The decompressed image as an pyopencl array.
        :rtype: pyopencl.array
        """
        assert self.dec_size is not None, \
            "dec_size is a mandatory ByteOffset init argument for decompression"

        events = []
        with self.sem:
            len_raw = numpy.int32(len(raw))
            if len_raw > self.padded_raw_size:
                wg = self.block_size
                self.raw_size = int(len(raw))
                self.padded_raw_size = (self.raw_size + wg - 1) & ~(wg - 1)
                logger.info("increase raw buffer size to %s", self.padded_raw_size)
                buffers = {
                           "raw": pyopencl.array.empty(self.queue, self.padded_raw_size, dtype=numpy.int8),
                           "mask": pyopencl.array.empty(self.queue, self.padded_raw_size, dtype=numpy.int32),
                           "exceptions": pyopencl.array.empty(self.queue, self.padded_raw_size, dtype=numpy.int32),
                           "values": pyopencl.array.empty(self.queue, self.padded_raw_size, dtype=numpy.int32),
                          }
                self.cl_mem.update(buffers)
            else:
                wg = self.block_size

            evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["raw"].data,
                                        raw,
                                        is_blocking=False)
            events.append(EventDescription("copy raw H -> D", evt))
            evt = self.kernels.fill_int_mem(self.queue, (self.padded_raw_size,), (wg,),
                                            self.cl_mem["mask"].data,
                                            numpy.int32(self.padded_raw_size),
                                            numpy.int32(0),
                                            numpy.int32(0))
            events.append(EventDescription("memset mask", evt))
            evt = self.kernels.fill_int_mem(self.queue, (1,), (1,),
                                            self.cl_mem["counter"].data,
                                            numpy.int32(1),
                                            numpy.int32(0),
                                            numpy.int32(0))
            events.append(EventDescription("memset counter", evt))
            evt = self.kernels.mark_exceptions(self.queue, (self.padded_raw_size,), (wg,),
                                               self.cl_mem["raw"].data,
                                               len_raw,
                                               numpy.int32(self.raw_size),
                                               self.cl_mem["mask"].data,
                                               self.cl_mem["values"].data,
                                               self.cl_mem["counter"].data,
                                               self.cl_mem["exceptions"].data)
            events.append(EventDescription("mark exceptions", evt))
            nb_exceptions = numpy.empty(1, dtype=numpy.int32)
            evt = pyopencl.enqueue_copy(self.queue, nb_exceptions, self.cl_mem["counter"].data,
                                        is_blocking=False)
            events.append(EventDescription("copy counter D -> H", evt))
            evt.wait()
            nbexc = int(nb_exceptions[0])
            if nbexc == 0:
                logger.info("nbexc %i", nbexc)
            else:
                evt = self.kernels.treat_exceptions(self.queue, (nbexc,), (1,),
                                                    self.cl_mem["raw"].data,
                                                    len_raw,
                                                    self.cl_mem["mask"].data,
                                                    self.cl_mem["exceptions"].data,
                                                    self.cl_mem["values"].data
                                                    )
                events.append(EventDescription("treat_exceptions", evt))

            #self.cl_mem["copy_values"] = self.cl_mem["values"].copy()
            #self.cl_mem["copy_mask"] = self.cl_mem["mask"].copy()
            evt = self.kernels.scan(self.cl_mem["values"],
                                    self.cl_mem["mask"],
                                    queue=self.queue,
                                    size=int(len_raw),
                                    wait_for=(evt,))
            events.append(EventDescription("double scan", evt))
            #evt.wait()
            if out is not None:
                if out.dtype == numpy.float32:
                    copy_results = self.kernels.copy_result_float
                else:
                    copy_results = self.kernels.copy_result_int
            else:
                if as_float:
                    out = self.cl_mem["data_float"]
                    copy_results = self.kernels.copy_result_float
                else:
                    out = self.cl_mem["data_int"]
                    copy_results = self.kernels.copy_result_int
            evt = copy_results(self.queue, (self.padded_raw_size,), (wg,),
                               self.cl_mem["values"].data,
                               self.cl_mem["mask"].data,
                               len_raw,
                               self.dec_size,
                               out.data
                               )
            events.append(EventDescription("copy_results", evt))
            #evt.wait()
            if self.profile:
                self.events += events
        return out

    __call__ = decode

    def _init_compression_scan(self):
        """Initialize CBF compression scan kernels"""
        preamble = """
        int compressed_size(int diff) {
            int abs_diff = abs(diff);

            if (abs_diff < 128) {
                return 1;
            }
            else if (abs_diff < 32768) {
                return 3;
            }
            else {
                return 7;
            }
        }

        void write(const int index,
                   const int diff,
                   global char *output) {
            int abs_diff = abs(diff);

            if (abs_diff < 128) {
                output[index] = (char) diff;
            }
            else if (abs_diff < 32768) {
                output[index] = -128;
                output[index + 1] = (char) (diff >> 0);
                output[index + 2] = (char) (diff >> 8);
            }
            else {
                output[index] = -128;
                output[index + 1] = 0;
                output[index + 2] = -128;
                output[index + 3] = (char) (diff >> 0);
                output[index + 4] = (char) (diff >> 8);
                output[index + 5] = (char) (diff >> 16);
                output[index + 6] = (char) (diff >> 24);
            }
        }
        """
        arguments = "__global const int *data, __global char *compressed, __global int *size"
        input_expr = "compressed_size((i == 0) ? data[0] : (data[i] - data[i - 1]))"
        scan_expr = "a+b"
        neutral = "0"
        output_statement = """
        if (prev_item == 0) { // 1st thread store compressed data size
            size[0] = last_item;
        }
        write(prev_item, (i == 0) ? data[0] : (data[i] - data[i - 1]), compressed);
        """

        if self.block_size >= 64:
            knl = GenericScanKernel(self.ctx,
                                    dtype=numpy.int32,
                                    preamble=preamble,
                                    arguments=arguments,
                                    input_expr=input_expr,
                                    scan_expr=scan_expr,
                                    neutral=neutral,
                                    output_statement=output_statement)
        else:  # MacOS on CPU
            knl = GenericDebugScanKernel(self.ctx,
                                         dtype=numpy.int32,
                                         preamble=preamble,
                                         arguments=arguments,
                                         input_expr=input_expr,
                                         scan_expr=scan_expr,
                                         neutral=neutral,
                                         output_statement=output_statement)
        return knl

    def encode(self, data, out=None):
        """Compress data to CBF.

        :param data: The data to compress as a numpy array
                     (or a pyopencl Array) of int32.
        :type data: Union[numpy.ndarray, pyopencl.array.Array]
        :param pyopencl.array out:
            pyopencl array of int8 in which to store the result.
            The array should be large enough to store the compressed data.
        :return: The compressed data as a pyopencl array.
                 If out is provided, this array shares the backing buffer,
                 but has the exact size of the compressed data and the queue
                 of the ByteOffset instance.
        :rtype: pyopencl.array
        :raises ValueError: if out array is not large enough
        """

        events = []
        with self.sem:
            if isinstance(data, pyopencl.array.Array):
                d_data = data  # Uses provided array

            else:  # Copy data to device
                data = numpy.ascontiguousarray(data, dtype=numpy.int32).ravel()

                # Make sure data array exists and is large enough
                if ("data_input" not in self.cl_mem or
                        self.cl_mem["data_input"].size < data.size):
                    logger.info("increase data input buffer size to %s", data.size)
                    self.cl_mem.update({
                        "data_input": pyopencl.array.empty(self.queue,
                                                           data.size,
                                                           dtype=numpy.int32)})
                d_data = self.cl_mem["data_input"]

                evt = pyopencl.enqueue_copy(
                    self.queue, d_data.data, data, is_blocking=False)
                events.append(EventDescription("copy data H -> D", evt))

            # Make sure compressed array exists and is large enough
            compressed_size = d_data.size * 7
            if ("compressed" not in self.cl_mem or
                    self.cl_mem["compressed"].size < compressed_size):
                logger.info("increase compressed buffer size to %s", compressed_size)
                self.cl_mem.update({
                    "compressed": pyopencl.array.empty(self.queue,
                                                       compressed_size,
                                                       dtype=numpy.int8)})
            d_compressed = self.cl_mem["compressed"]
            d_size = self.cl_mem["counter"]  # Shared with decompression

            evt = self.kernels.compression_scan(d_data, d_compressed, d_size)
            events.append(EventDescription("compression scan", evt))
            byte_count = int(d_size.get()[0])

            if out is None:
                # Create out array from a sub-region of the compressed buffer
                out = pyopencl.array.Array(
                    self.queue,
                    shape=(byte_count,),
                    dtype=numpy.int8,
                    allocator=functools.partial(
                        d_compressed.base_data.get_sub_region,
                        d_compressed.offset))

            elif out.size < byte_count:
                raise ValueError(
                    "Provided output buffer is not large enough: "
                    "requires %d bytes, got %d" % (byte_count, out.size))

            else:  # out.size >= byte_count
                # Create an array with a sub-region of out and this class queue
                out = pyopencl.array.Array(
                    self.queue,
                    shape=(byte_count,),
                    dtype=numpy.int8,
                    allocator=functools.partial(out.base_data.get_sub_region,
                                                out.offset))

                evt = pyopencl.enqueue_copy(self.queue, out.data, d_compressed.data,
                                            byte_count=byte_count)
                events.append(
                    EventDescription("copy D -> D: internal -> out", evt))

            if self.profile:
                self.events += events

        return out

    def encode_to_bytes(self, data):
        """Compresses data to CBF and returns compressed data as bytes.

        Usage:

        Provided an image (`image`) stored as a numpy array of int32,
        first, create a byte offset compression/decompression object:

        >>> from silx.opencl.codec.byte_offset import ByteOffset
        >>> byte_offset_codec = ByteOffset()

        Then, compress an image into bytes:

        >>> compressed = byte_offset_codec.encode_to_bytes(image)

        :param data: The data to compress as a numpy array
                     (or a pyopencl Array) of int32.
        :type data: Union[numpy.ndarray, pyopencl.array.Array]
        :return: The compressed data as bytes.
        :rtype: bytes
        """
        compressed_array = self.encode(data)
        return compressed_array.get().tobytes()



#/*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "22/06/2016"

cimport cython

cdef extern from "filters.h":
    void snip1d(double *data,
                int size,
                int width)

    void snip2d(double *data,
                int nrows,
                int ncolumns,
                int width)

    void snip3d(double *data,
                int nx,
                int ny,
                int nz,
                int width)

    int strip(double* input,
              long len_input,
              double c,
              long niter,
              int deltai,
              long* anchors,
              long len_anchors,
              double* output)

    int SavitskyGolay(double* input,
                      long len_input,
                      int npoints,
                      double* output)

    void smooth1d(double *data,
                  int size)

    void smooth2d(double *data,
                  int size0,
                  int size1)

    void smooth3d(double *data,
                  int size0,
                  int size1,
                  int size2)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
import numpy as np
from pkg_resources import parse_version


def check_version(package, required_version):
    """
    Check whether a given package version is superior or equal to required_version.
    """
    try:
        ver = getattr(package, "__version__")
    except AttributeError:
        try:
            ver = getattr(package, "version")
        except Exception:
            return False
    req_v = parse_version(required_version)
    ver_v = parse_version(ver)
    return ver_v >= req_v


class BaseFFT(object):
    """
    Base class for all FFT backends.
    """
    def __init__(self, **kwargs):
        self.__get_args(**kwargs)

        if self.shape is None and self.dtype is None and self.template is None:
            raise ValueError("Please provide either (shape and dtype) or template")
        if self.template is not None:
            self.shape = self.template.shape
            self.dtype = self.template.dtype
        self.user_data = self.template
        self.data_allocated = False
        self.__calc_axes()
        self.__set_dtypes()
        self.__calc_shape()

    def __get_args(self, **kwargs):
        expected_args = {
            "shape": None,
            "dtype": None,
            "template": None,
            "shape_out": None,
            "axes": None,
            "normalize": "rescale",
        }
        for arg_name, default_val in expected_args.items():
            if arg_name not in kwargs:
                # Base class was not instantiated properly
                raise ValueError("Please provide argument %s" % arg_name)
            setattr(self, arg_name, default_val)
        for arg_name, arg_val in kwargs.items():
            setattr(self, arg_name, arg_val)

    def __set_dtypes(self):
        dtypes_mapping = {
            np.dtype("float32"): np.complex64,
            np.dtype("float64"): np.complex128,
            np.dtype("complex64"): np.complex64,
            np.dtype("complex128"): np.complex128
        }
        dp = {
            np.dtype("float32"): np.float64,
            np.dtype("complex64"): np.complex128
        }
        self.dtype_in = np.dtype(self.dtype)
        if self.dtype_in not in dtypes_mapping:
            raise ValueError("Invalid input data type: got %s" %
                self.dtype_in
            )
        self.dtype_out = dtypes_mapping[self.dtype_in]

    def __calc_shape(self):
        # TODO allow for C2C even for real input data (?)
        if self.dtype_in in [np.float32, np.float64]:
            last_dim = self.shape[-1]//2 + 1
            # FFTW convention
            self.shape_out = self.shape[:-1] + (self.shape[-1]//2 + 1,)
        else:
            self.shape_out = self.shape

    def __calc_axes(self):
        default_axes = tuple(range(len(self.shape)))
        if self.axes is None:
            self.axes = default_axes
            self.user_axes = None
        else:
            self.user_axes = self.axes
            # Handle possibly negative axes
            self.axes = tuple(np.array(default_axes)[np.array(self.user_axes)])

    def _allocate(self, shape, dtype):
        raise ValueError("This should be implemented by back-end FFT")

    def set_data(self, dst, src, shape, dtype, copy=True):
        raise ValueError("This should be implemented by back-end FFT")

    def allocate_arrays(self):
        if not(self.data_allocated):
            self.data_in = self._allocate(self.shape, self.dtype_in)
            self.data_out = self._allocate(self.shape_out, self.dtype_out)
            self.data_allocated = True

    def set_input_data(self, data, copy=True):
        if data is None:
            return self.data_in
        else:
            return self.set_data(self.data_in, data, self.shape, self.dtype_in, copy=copy, name="data_in")

    def set_output_data(self, data, copy=True):
        if data is None:
            return self.data_out
        else:
            return self.set_data(self.data_out, data, self.shape_out, self.dtype_out, copy=copy, name="data_out")

    def fft(self, array, **kwargs):
        raise ValueError("This should be implemented by back-end FFT")

    def ifft(self, array, **kwargs):
        raise ValueError("This should be implemented by back-end FFT")



# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test of actions integrated in the plot window"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "09/11/2018"


import pytest
import weakref

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.PlotWindow import PlotWindow

import numpy


@pytest.fixture
def colormap1():
    colormap = Colormap(name='gray',
                        vmin=10.0, vmax=20.0,
                        normalization='linear')
    yield colormap


@pytest.fixture
def colormap2():
    colormap = Colormap(name='red',
                        vmin=10.0, vmax=20.0,
                        normalization='linear')
    yield colormap


@pytest.fixture
def plot(qapp):
    plot = PlotWindow()
    plot.setAttribute(qt.Qt.WA_DeleteOnClose)
    yield weakref.proxy(plot)
    plot.close()
    qapp.processEvents()


def test_action_active_colormap(qapp_utils, plot, colormap1, colormap2):
    plot.getColormapAction()._actionTriggered(checked=True)
    colormapDialog = plot.getColormapAction()._dialog

    defaultColormap = plot.getDefaultColormap()
    assert colormapDialog.getColormap() is defaultColormap

    plot.addImage(data=numpy.random.rand(10, 10), legend='img1',
                  origin=(0, 0),
                  colormap=colormap1)
    plot.setActiveImage('img1')
    assert colormapDialog.getColormap() is colormap1

    plot.addImage(data=numpy.random.rand(10, 10), legend='img2',
                  origin=(0, 0), colormap=colormap2)
    plot.addImage(data=numpy.random.rand(10, 10), legend='img3',
                  origin=(0, 0))

    plot.setActiveImage('img3')
    assert colormapDialog.getColormap() is defaultColormap
    plot.getActiveImage().setColormap(colormap2)
    assert colormapDialog.getColormap() is colormap2

    plot.remove('img2')
    plot.remove('img3')
    plot.remove('img1')
    assert colormapDialog.getColormap() is defaultColormap


def test_action_show_hide_colormap_dialog(qapp_utils, plot, colormap1):
    plot.getColormapAction()._actionTriggered(checked=True)
    colormapDialog = plot.getColormapAction()._dialog

    plot.getColormapAction()._actionTriggered(checked=False)
    assert not plot.getColormapAction().isChecked()
    plot.getColormapAction()._actionTriggered(checked=True)
    assert plot.getColormapAction().isChecked()
    plot.addImage(data=numpy.random.rand(10, 10), legend='img1',
                  origin=(0, 0), colormap=colormap1)
    colormap1.setName('red')
    plot.getColormapAction()._actionTriggered()
    colormap1.setName('blue')
    colormapDialog.close()
    assert not plot.getColormapAction().isChecked()
